import os
import time

# import multiprocessing
from pathlib import Path
from enum import Enum, IntFlag, auto
from dataclasses import dataclass
from typing import Callable, Optional, List
from concurrent.futures import ThreadPoolExecutor

from seal5.logging import Logger
from seal5.settings import Seal5Settings

logger = Logger("passes")


# DEFAULT_NUM_THREADS = multiprocessing.cpu_count()
DEFAULT_NUM_THREADS = 1
NUM_THREADS = int(os.environ.get("SEAL5_NUM_THREADS", DEFAULT_NUM_THREADS))


class PassFormat(IntFlag):
    NONE = auto()
    CDSL = auto()
    YAML = auto()
    M2ISAR = auto()
    SEAL5 = auto()
    LLVMIR = auto()
    GMIR = auto()
    SRC = auto()


class PassType(Enum):
    TRANSFORM = auto()
    GENERATE = auto()


class PassStatus(Enum):
    CREATED = auto()
    RUNNING = auto()
    FAILED = auto()
    COMPLETED = auto()
    SKIPPED = auto()


class PassScope(Enum):
    GLOBAL = auto()
    MODEL = auto()
    SET = auto()  # not supported yet
    INSTR = auto()  # not supported yet


@dataclass
class PassResult:
    metrics: Optional[dict] = None
    outputs: Optional[List[Path]] = None


def check_filter(name, skip, only):
    if skip is None:
        skip = []
    if only is None:
        only = []
    if not ((name in only or len(only) == 0) and (name not in skip or len(skip) == 0)):
        return True
    return False


class Seal5Pass:
    def __init__(self, name, pass_type, pass_scope, handler, fmt=PassFormat.NONE, order=-1, options=None):
        self.name: str = name
        self.pass_type: PassType = pass_type
        self.pass_scope: PassScope = pass_scope
        self.handler: Callable = handler
        self.fmt: PassFormat = fmt
        self.status = PassStatus.CREATED
        self.order: int = order  # not supported yet
        self.options: Optional[dict] = options
        self.metrics: dict = {}

    def __repr__(self):
        return f"Seal5Pass({self.name}, {self.pass_type}, {self.pass_scope})"

    @property
    def is_pending(self):
        return self.status in [PassStatus.CREATED, PassStatus.SKIPPED]

    def skip(self):
        self.status = PassStatus.SKIPPED

    def run(self, inputs: List[str], *_args, settings: Optional[Seal5Settings] = None, **kwargs):
        logger.debug("Running pass: %s", self)
        self.status = PassStatus.RUNNING
        assert settings is not None
        passes_settings = settings.passes
        assert passes_settings is not None
        self.metrics["models"] = []
        try:
            kwargs_ = {**kwargs}
            if self.options:
                kwargs_.update(self.options)
            start = time.time()
            parent = kwargs.get("parent", None)
            if parent:
                parallel = parent.parallel
            else:
                parallel = 1
            if self.pass_scope == PassScope.SET:
                with ThreadPoolExecutor(max_workers=parallel) as executor:
                    futures = []
                    passes_settings_ = passes_settings
                    for input_model in inputs:
                        model_settings = settings.models[input_model]
                        model_passes_settings = model_settings.passes
                        if model_passes_settings is not None:
                            passes_settings_ = passes_settings_.merge(model_passes_settings)
                        if check_filter(self.name, passes_settings_.skip, passes_settings_.only):
                            logger.info("Skipped pass %s for model %s", self.name, input_model)
                            continue
                        passes_settings__ = passes_settings_
                        for ext_name, ext_settings in model_settings.extensions.items():
                            ext_passes_settings = model_settings.passes
                            if ext_passes_settings is not None:
                                passes_settings__ = passes_settings_.merge(ext_passes_settings)
                            if check_filter(self.name, passes_settings__.skip, passes_settings__.only):
                                logger.info("Skipped pass %s for extension %s", self.name, ext_name)
                                continue
                            kwargs__ = kwargs_.copy()
                            if passes_settings__.overrides:
                                overrides = passes_settings__.overrides.get(self.name)
                                if overrides:
                                    kwargs__.update(overrides)
                            future = executor.submit(self.handler, input_model, ext_name, settings=settings, **kwargs__)
                            futures.append(future)
                    results = []
                    for i, future in enumerate(futures):
                        result = future.result()
                        input_model = inputs[i]
                        if result:
                            metrics = result.metrics
                            if metrics:
                                self.metrics["models"].append({input_model: metrics})
                        results.append(result)
                    # TODO: check results (metrics?)
            elif self.pass_scope == PassScope.MODEL:

                with ThreadPoolExecutor(max_workers=parallel) as executor:
                    futures = []
                    passes_settings_ = passes_settings
                    for input_model in inputs:
                        model_settings = settings.models.get(input_model)
                        model_passes_settings = model_settings.passes if model_settings is not None else None
                        kwargs__ = kwargs_.copy()
                        if model_passes_settings is not None:
                            passes_settings_ = passes_settings_.merge(model_passes_settings)
                        if check_filter(self.name, passes_settings_.skip, passes_settings_.only):
                            logger.info("Skipped pass %s for model %s", self.name, input_model)
                            continue
                        if passes_settings_.overrides:
                            overrides = passes_settings_.overrides.get(self.name)
                            if overrides:
                                kwargs__.update(overrides)
                        future = executor.submit(self.handler, input_model, settings=settings, **kwargs__)
                        futures.append(future)
                    results = []
                    for i, future in enumerate(futures):
                        result = future.result()
                        input_model = inputs[i]
                        if result:
                            metrics = result.metrics
                            if metrics:
                                self.metrics["models"].append({input_model: metrics})
                        results.append(result)
                    # TODO: check results (metrics?)
            elif self.pass_scope == PassScope.GLOBAL:
                input_model = None
                result = self.handler(input_model, settings=settings, **kwargs_)
                if result:
                    metrics = result.metrics
                    if metrics:
                        self.metrics["models"].append({input_model: metrics})
            else:
                raise NotImplementedError
            end = time.time()
            diff = end - start
            self.status = PassStatus.COMPLETED
            self.metrics["start"] = start
            self.metrics["end"] = end
            self.metrics["time_s"] = diff
        except Exception as e:
            self.status = PassStatus.FAILED
            raise e
        return PassResult(metrics=self.metrics)


class PassManager:
    def __init__(
        self,
        name: str,
        pass_list: List[Seal5Pass],
        skip: Optional[List[str]] = None,
        only: Optional[List[str]] = None,
        parent: Optional["PassManager"] = None,
        parallel: Optional[int] = None,
    ):
        self.name = name
        self.pass_list = pass_list
        self.skip = skip if skip is not None else (parent.skip if parent else [])
        self.only = only if only is not None else (parent.only if parent else [])
        self.parallel = parallel if parallel is not None else (parent.parallel if parent else NUM_THREADS)
        self.metrics: dict = {}
        self.open: bool = False

    @property
    def size(self):
        return len(self.pass_list)

    def __enter__(self):
        assert not self.open
        self.open = True
        logger.debug("Processing %d passes...", self.size)
        return self

    def __exit__(self, *exc):
        self.open = False
        logger.debug("Done.")
        # return False

    def run(
        self,
        input_models: List[str],
        settings: Optional[Seal5Settings] = None,
        env: Optional[dict] = None,
        verbose: bool = False,
    ):
        assert self.open, "PassManager needs context"
        start = time.time()
        self.metrics["passes"] = []
        # passes_settings = settings.passes
        # assert passes_settings is not None
        # assert passes_settings.per_model is not None

        for pass_ in self.pass_list:
            # input_models_ = []
            # for model_name in input_models:
            #    overrides = passes_settings.per_model.get(model_name, None)
            #    if overrides:
            if check_filter(pass_.name, self.skip, self.only):
                pass_.skip()
                continue
            assert pass_.is_pending, f"Pass {pass_.name} is not pending"
            result = pass_.run(input_models, settings=settings, env=env, verbose=verbose, parent=self)
            if result:
                metrics = result.metrics
                if metrics:
                    self.metrics["passes"].append({pass_.name: metrics})
        end = time.time()
        diff = end - start
        self.metrics["start"] = start
        self.metrics["end"] = end
        self.metrics["time_s"] = diff
        return PassResult(metrics=self.metrics)


def filter_passes(
    passes: List[Seal5Pass],
    pass_name: Optional[str] = None,
    pass_type: Optional[PassType] = None,
    pass_scope: Optional[PassScope] = None,
):
    # TODO: support regex patters in name
    ret = []
    for pass_ in passes:
        if pass_name is not None:
            if pass_.name != pass_name:
                continue
        if pass_type is not None:
            if pass_.pass_type != pass_type:
                continue
        if pass_scope is not None:
            if pass_.pass_scope != pass_scope:
                continue
        ret.append(pass_)
    return ret



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

import argparse
import itertools
import logging
import pathlib
import pickle
import sys

from m2isar import M2Error, M2SyntaxError
from m2isar.metamodel import M2_METAMODEL_VERSION, M2Model, behav, patch_model
from m2isar.metamodel.code_info import CodeInfoBase
from . import expr_interpreter
from .architecture_model_builder import ArchitectureModelBuilder
from .behavior_model_builder import BehaviorModelBuilder
from .importer import recursive_import
from .load_order import LoadOrder
from .utils import make_parser
from seal5.logging import Logger


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="The CoreDSL file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)

    args = parser.parse_args()

    # app_dir = pathlib.Path(__file__).parent.resolve()

    logger = Logger("frontends.parser")
    logger.setLevel(getattr(logging, args.log.upper()))

    top_level = pathlib.Path(args.top_level)
    abs_top_level = top_level.resolve()
    search_path = abs_top_level.parent

    parser = make_parser(abs_top_level)

    try:
        logger.info("parsing top level")
        tree = parser.description_content()

        recursive_import(tree, search_path)
    except M2SyntaxError as e:
        logger.critical("Error during parsing: %s", e)
        sys.exit(1)

    logger.info("reading instruction load order")
    lo = LoadOrder()
    try:
        sets = lo.visit(tree)
    except M2Error as e:
        logger.critical("Error during load order building: %s", e)
        sys.exit(1)

    if args.output is None:
        model_path = search_path.joinpath("gen_model")
    else:
        model_path = pathlib.Path(args.output)
    model_path.mkdir(exist_ok=True)

    temp_save = {}

    patch_model(expr_interpreter)

    for set_name, set_def in sets.items():
        logger.info(f"building architecture model for set {set_name}")
        try:
            arch_builder = ArchitectureModelBuilder()
            s = arch_builder.visit(set_def)
            if not isinstance(s, list):
                s = [s]
            # print("s", s)
        except M2Error as e:
            logger.critical("Error building architecture model of core", e)

        for orig, overwritten in arch_builder._overwritten_instrs:
            logger.warning(
                "instr %s from extension %s was overwritten by %s from %s",
                orig.name,
                orig.ext_name,
                overwritten.name,
                overwritten.ext_name,
            )

        temp_save[set_name] = (s, arch_builder)

        sets[set_name] = s[-1]

    for set_name, set_def in sets.items():
        logger.info("building behavior model for set %s", set_name)
        # print("set", set_name, set_def, dir(set_def))

        warned_fns = set()

        logger.debug("checking core constants")
        unassigned_const = False
        for const in set_def.constants.values():
            # print("const", const)
            if const.value is None:
                pass
                # if const.name == "XLEN":
                # 	if "32" in set_name:
                # 		const.value = 32
                # 	elif "64" in set_name:
                # 		const.value = 64
                # 	continue
                # logger.critical("constant %s in set %s has no value assigned!", const.name, set_name)
                # unassigned_const = True
                # sys.exit(-1)
        if unassigned_const:
            sys.exit(-1)

        logger.debug("evaluating set parameters")

        for const_def in set_def.constants.values():
            const_def._value = const_def.value

        for mem_def in itertools.chain(set_def.memories.values(), set_def.memory_aliases.values()):
            mem_def._size = mem_def.size
            mem_def.range._lower_base = mem_def.range.lower_base
            mem_def.range._upper_base = mem_def.range.upper_base

            for attr_name, attr_ops in mem_def.attributes.items():
                ops = []
                for attr_op in attr_ops:
                    try:
                        behav_builder = BehaviorModelBuilder(
                            set_def.constants,
                            set_def.memories,
                            set_def.memory_aliases,
                            {},
                            set_def.functions,
                            warned_fns,
                        )
                        op = behav_builder.visit(attr_op)
                        ops.append(op)
                    except M2Error as e:
                        logger.critical(
                            'error processing attribute "%s" of memory "%s": %s', attr_name, mem_def.name, e
                        )
                        sys.exit(1)

                mem_def.attributes[attr_name] = ops

        for fn_def in set_def.functions.values():
            if isinstance(fn_def.operation, behav.Operation) and not fn_def.extern:
                raise M2SyntaxError(f"non-extern function {fn_def.name} has no body")

            fn_def._size = fn_def.size
            for fn_arg in fn_def.args.values():
                fn_arg._size = fn_arg.size
                fn_arg._width = fn_arg.width

        logger.debug("generating function behavior")

        for fn_name, fn_def in set_def.functions.items():
            logger.debug("generating function %s", fn_name)
            logger.debug("generating attributes")

            for attr_name, attr_ops in fn_def.attributes.items():
                ops = []
                for attr_op in attr_ops:
                    try:
                        behav_builder = BehaviorModelBuilder(
                            set_def.constants,
                            set_def.memories,
                            set_def.memory_aliases,
                            fn_def.args,
                            set_def.functions,
                            warned_fns,
                        )
                        op = behav_builder.visit(attr_op)
                        ops.append(op)
                    except M2Error as e:
                        logger.critical(
                            'error processing attribute "%s" of function "%s": %s', attr_name, fn_def.name, e
                        )
                        sys.exit(1)

                fn_def.attributes[attr_name] = ops

            behav_builder = BehaviorModelBuilder(
                set_def.constants, set_def.memories, set_def.memory_aliases, fn_def.args, set_def.functions, warned_fns
            )

            if not isinstance(fn_def.operation, behav.Operation):
                try:
                    op = behav_builder.visit(fn_def.operation)
                except M2Error as e:
                    logger.critical("Error building behavior for function %s: %s", fn_name, e)
                    sys.exit()

                fn_def.scalars = behav_builder._scalars

                if isinstance(op, list):
                    fn_def.operation = behav.Operation(op)
                else:
                    fn_def.operation = behav.Operation([op])

        logger.debug("generating always blocks")

        always_block_statements = []

        arch_builder = temp_save[set_name][1]
        for block_def in arch_builder._always_blocks.values():
            logger.debug("generating always block %s", block_def.name)
            logger.debug("generating attributes")

            for attr_name, attr_ops in block_def.attributes.items():
                ops = []
                for attr_op in attr_ops:
                    try:
                        behav_builder = BehaviorModelBuilder(
                            set_def.constants,
                            set_def.memories,
                            set_def.memory_aliases,
                            {},
                            set_def.functions,
                            warned_fns,
                        )
                        op = behav_builder.visit(attr_op)
                        ops.append(op)
                    except M2Error as e:
                        logger.critical(
                            'error processing attribute "%s" of instruction "%s": %s', attr_name, block_def.name, e
                        )
                        sys.exit(1)

                block_def.attributes[attr_name] = ops

            behav_builder = BehaviorModelBuilder(
                set_def.constants, set_def.memories, set_def.memory_aliases, {}, set_def.functions, warned_fns
            )

            try:
                op = behav_builder.visit(block_def.operation)
            except M2Error as e:
                logger.critical("error building behavior for always block %s: %s", block_def.name, e)
                sys.exit(1)

            always_block_statements.append(op)

        logger.debug("generating instruction behavior")

        for instr_def in set_def.instructions.values():
            logger.debug("generating instruction %s", instr_def.name)
            logger.debug("generating attributes")

            for attr_name, attr_ops in instr_def.attributes.items():
                ops = []
                for attr_op in attr_ops:
                    try:
                        behav_builder = BehaviorModelBuilder(
                            set_def.constants,
                            set_def.memories,
                            set_def.memory_aliases,
                            instr_def.fields,
                            set_def.functions,
                            warned_fns,
                        )
                        op = behav_builder.visit(attr_op)
                        ops.append(op)
                    except M2Error as e:
                        logger.critical(
                            'error processing attribute "%s" of instruction "%s": %s', attr_name, instr_def.name, e
                        )
                        sys.exit(1)

                instr_def.attributes[attr_name] = ops

            behav_builder = BehaviorModelBuilder(
                set_def.constants,
                set_def.memories,
                set_def.memory_aliases,
                instr_def.fields,
                set_def.functions,
                warned_fns,
            )

            try:
                op = behav_builder.visit(instr_def.operation)
            except M2Error as e:
                logger.critical(
                    "error building behavior for instruction %s::%s: %s", instr_def.ext_name, instr_def.name, e
                )
                sys.exit(1)

            instr_def.scalars = behav_builder._scalars

            if isinstance(op, list):
                op = behav.Operation(op)
            else:
                op = behav.Operation([op])

            # pc_inc = behav.Assignment(
            # 	behav.NamedReference(set_def.pc_memory),
            # 	behav.BinaryOperation(
            # 		behav.NamedReference(set_def.pc_memory),
            # 		behav.Operator("+"),
            # 		behav.IntLiteral(int(instr_def.size/8))
            # 	)
            # )

            # op.statements.insert(0, pc_inc)
            op.statements = always_block_statements + op.statements
            instr_def.operation = op

    logger.info("dumping model")
    with open(model_path / (abs_top_level.stem + ".m2isarmodel"), "wb") as f:
        model_obj = M2Model(M2_METAMODEL_VERSION, {}, sets, CodeInfoBase.database)
        pickle.dump(model_obj, f)


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Status report writer for Seal5."""

import argparse
import logging
import pathlib
from collections import defaultdict

import yaml
import pandas as pd

from seal5.settings import Seal5Settings

logger = logging.getLogger("status_writer")


def main():
    """Main app entrypoint."""

    # read command line args
    parser = argparse.ArgumentParser()
    # parser.add_argument("top_level", nargs="+", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--fmt", type=str, choices=["auto", "csv", "pkl", "md"], default="auto")
    parser.add_argument("--yaml", type=str, default=None)
    parser.add_argument("--agg", action="store_true")
    parser.add_argument("--print-df", action="store_true")
    args = parser.parse_args()

    # initialize logging
    logging.basicConfig(level=getattr(logging, args.log.upper()))

    # if args.output is None:
    #     assert top_level.suffix in [".m2isarmodel", ".seal5model"], "Can not infer model type from file extension."
    #     raise NotImplementedError

    #     # out_path = top_level.parent / (top_level.stem + ".core_desc")
    # else:
    assert args.output is not None
    out_path = pathlib.Path(args.output)

    assert args.yaml is not None
    assert pathlib.Path(args.yaml).is_file()
    settings = Seal5Settings.from_yaml_file(args.yaml)
    meta_dir = pathlib.Path(settings.meta_dir)
    # TODOL combine stats automnatically via metrics?
    temp_dir = meta_dir / "temp"
    assert temp_dir.is_dir()
    # all_data = {}
    all_data = []
    agg_data = defaultdict(int)
    for p in temp_dir.rglob("**/*.td.stats"):
        instr_name = p.name.split(".td", 1)[0]
        # print("instr_name", instr_name)
        with open(p, "r") as f:
            data = yaml.safe_load(f)
        assert "pattern-gen" in data
        data = data["pattern-gen"]
        # print("data", data)
        # all_data[instr_name] = data
        all_data.append({"instr_name": instr_name, **data})
        for key, val in data.items():
            agg_data[key] += val

    all_df = pd.DataFrame(all_data)
    agg_df = pd.Series(agg_data).to_frame().reset_index().rename(columns={"index": "stat", 0: "count"})

    df = agg_df if args.agg else all_df
    if args.print_df:
        print(df)

    fmt = args.fmt
    if fmt == "auto":
        fmt = out_path.suffix
        assert len(fmt) > 1
        fmt = fmt[1:].lower()

    if fmt == "csv":
        df.to_csv(out_path, index=False)
    elif fmt == "pkl":
        df.to_pickle(out_path)
    elif fmt == "md":
        df.to_markdown(out_path, index=False)
    else:
        raise ValueError(f"Unsupported fmt: {fmt}")


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Clean M2-ISA-R/Seal5 metamodel to .core_desc file."""

import argparse
import logging
import pathlib

import pandas as pd

from mako.template import Template

from seal5.index import NamedPatch, TestFile, write_index_yaml
from seal5.settings import ExtensionsSettings, LLVMSettings
from seal5.model_utils import load_model

from .templates import template_dir


from seal5.logging import Logger

logger = Logger("backends.riscv_features")


def gen_riscv_features_str(
    name: str, ext_settings: ExtensionsSettings, llvm_settings: LLVMSettings, all_sets, generate_tests
):
    """Generate features string for LLVM patch."""
    # requires = ext_settings.requires
    implies = ext_settings.requires
    feature = ext_settings.get_feature(name=name)
    arch_ = ext_settings.get_arch(name=name)
    description = ext_settings.get_description(name=name)
    predicate = ext_settings.get_predicate(name=name)
    version = ext_settings.get_version()
    experimental = ext_settings.experimental
    vendor = ext_settings.vendor

    implied_features = set()
    implies_archs_with_ver = []
    if implies:
        for implied in implies:
            assert implied in all_sets
            implied_set_def = all_sets[implied]
            implied_ext_settings = implied_set_def.settings
            implied_feature = implied_ext_settings.get_predicate(name=implied)
            if generate_tests:
                implied_arch = implied_ext_settings.get_arch(name=implied)
                implied_version = implied_ext_settings.get_version()
                implied_version_str = str(implied_version).replace(".", "p")
                implied_arch_with_ver = implied_arch + implied_version_str
                implies_archs_with_ver.append(implied_arch_with_ver)
            # implied_features.add(f"Feature{implied_feature}")
            implied_features.add(f"Feature{implied_feature}")  # TODO: check missing Ext?

    legacy = True
    slim = False
    if llvm_settings:
        llvm_state = llvm_settings.state
        if llvm_state:
            llvm_version = llvm_state.version  # unused today, but needed very soon
            if llvm_version.major >= 19:
                legacy = False
                if llvm_version.major >= 20:
                    slim = True

    test_template_name = "test_riscv_attributes"
    if legacy:
        template_name = "riscv_features"
    else:
        if experimental:
            template_name = "riscv_features_experimental_new_slim" if slim else "riscv_features_experimental_new"
        else:
            template_name = "riscv_features_new_slim" if slim else "riscv_features_new"

    # TODO: make util!
    if not isinstance(version, str):
        assert isinstance(version, float)
        version = f"{version:.1f}"
    major, minor = list(map(int, version.split(".", 1)))

    content_template = Template(filename=str(template_dir / f"{template_name}.mako"))
    test_template = Template(filename=str(template_dir / f"{test_template_name}.mako"))
    if slim:
        # TODO: support experimental- prefix
        feature_lower = feature.lower()
        assert feature_lower == arch_, f"LLVM 20 requires matching arch and feature names ({feature_lower} vs. {arch_})"
        assert predicate == (f"Vendor{feature}" if vendor else f"StdExt{feature}")
    content_text = content_template.render(
        predicate=predicate,
        feature=feature,
        arch=arch_,
        description=description,
        major=major,
        minor=minor,
        implies="[" + ", ".join(implied_features) + "]",
    )
    test_files = {}
    if generate_tests:
        test_name = f"{arch_}.test-attrs.ll"
        xlen = ext_settings.riscv.xlen if ext_settings.riscv is not None else 32
        attrs_str = f"+{arch_}"
        label = arch_.upper()
        expected_archs = [f"{arch_}{major}p{minor}"]
        expected_archs += implies_archs_with_ver
        expected_archs_str = "_".join(expected_archs)
        expected = f"rv{xlen}i2p1_{expected_archs_str}"
        test_content = test_template.render(
            xlen=xlen,
            attrs_str=attrs_str,
            label=label,
            expected=expected,
        )
        test_files[test_name] = test_content
    return content_text + "\n", test_files


def main():
    """Main app entrypoint."""

    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--splitted", action="store_true", help="Split per set")
    parser.add_argument("--formats", action="store_true", help="Also generate instruction formats")
    parser.add_argument("--metrics", default=None, help="Output metrics to file")
    parser.add_argument("--index", default=None, help="Output index to file")
    parser.add_argument("--ext", type=str, default="td", help="Default file extension (if using --splitted)")
    parser.add_argument("--compat", action="store_true")
    parser.add_argument("--generate-tests", action="store_true")
    args = parser.parse_args()

    # initialize logging
    logger.setLevel(getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)
    out_path = pathlib.Path(args.output)

    model_obj = load_model(top_level, compat=args.compat)

    metrics = {
        "n_sets": 0,
        "n_skipped": 0,
        "n_failed": 0,
        "n_success": 0,
        "skipped_sets": [],
        "failed_sets": [],
        "success_sets": [],
    }
    # preprocess model
    # print("model", model)
    settings = model_obj.settings
    llvm_settings = None
    if settings.llvm:
        llvm_settings = settings.llvm
    artifacts = {}
    artifacts[None] = []  # used for global artifacts
    if not args.splitted:
        content = ""
        # errs = []
        for set_name, set_def in model_obj.sets.items():
            artifacts[set_name] = []
            metrics["n_sets"] += 1
            ext_settings = set_def.settings
            if ext_settings is None:
                metrics["n_skipped"] += 1
                metrics["skipped_sets"].append(set_name)
                continue
            metrics["n_success"] += 1
            metrics["success_sets"].append(set_name)
            content_, test_files_ = gen_riscv_features_str(
                set_name, ext_settings, llvm_settings, model_obj.sets, args.generate_tests
            )
            content += content_
            # test_files.update(test_files_)
            if args.generate_tests:
                for test_file, test_content in test_files_.items():
                    test_artifact = TestFile(
                        f"llvm/test/CodeGen/RISCV/seal5/generated/{test_file}", content=test_content
                    )
                    artifacts[set_name].append(test_artifact)
        content = content.rstrip()
        if len(content) > 0:
            with open(out_path, "w", encoding="utf-8") as f:
                f.write(content)
            riscv_features_patch = NamedPatch(
                "llvm/lib/Target/RISCV/RISCVFeatures.td", key="riscv_features", src_path=out_path
            )
            artifacts[None].append(riscv_features_patch)
    else:
        raise NotImplementedError
    if args.metrics:
        metrics_file = args.metrics
        metrics_file = args.metrics
        metrics_df = pd.DataFrame({key: [val] for key, val in metrics.items()})
        metrics_df.to_csv(metrics_file, index=False)
    if args.index:
        if sum(map(len, artifacts.values())) > 0:
            global_artifacts = artifacts.get(None, [])
            set_artifacts = {key: value for key, value in artifacts.items() if key is not None}
            index_file = args.index
            write_index_yaml(index_file, global_artifacts, set_artifacts, content=True)
        else:
            logger.warning("No patches generated. No index file will be written.")


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Viewer tool to visualize an M2-ISA-R model hierarchy."""

import argparse
import logging
import pathlib
import tkinter as tk
from collections import defaultdict
from tkinter import ttk
from anytree import Node, RenderTree

from m2isar.metamodel import arch, patch_model
from seal5.model_utils import load_model

from . import treegen
from .utils import TkTreeGenContext, TextTreeGenContext

from seal5.logging import Logger

logger = Logger("backends.viewer")


def sort_instruction(entry: "tuple[tuple[int, int], arch.Instruction]"):
    """Instruction sort key function. Sorts most restrictive encoding first."""
    (code, mask), _ = entry
    return bin(mask).count("1"), code
    # return code, bin(mask).count("1")


def main():
    """Main app entrypoint."""

    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel file containing the models to generate.")
    parser.add_argument("--text", "-t", action="store_true", help="TODO")
    parser.add_argument("--operation", "-o", action="store_true", help="TODO")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--compat", action="store_true")
    args = parser.parse_args()

    # initialize logging
    logger.setLevel(getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)
    abs_top_level = top_level.resolve()
    search_path = abs_top_level.parent.parent
    # model_fname = abs_top_level

    if abs_top_level.suffix == ".core_desc":
        logger.warning(".core_desc file passed as input. This is deprecated behavior, please change your scripts!")
        search_path = abs_top_level.parent
        model_path = search_path.joinpath("gen_model")

        if not model_path.exists():
            raise FileNotFoundError("Models not generated!")
        # model_fname = model_path / (abs_top_level.stem + ".m2isarmodel")

    output_base_path = search_path.joinpath("gen_output")
    output_base_path.mkdir(exist_ok=True)

    model_obj = load_model(top_level, compat=args.compat)

    # preprocess model
    # for core_name, core in models.items():
    # 	logger.info("preprocessing model %s", core_name)
    # 	process_functions(core)
    # 	process_instructions(core)
    # 	process_attributes(core)

    # load Ttk TreeView transformer functions
    patch_model(treegen)

    # create main Tk window
    if args.text:
        root = Node("Tree")
        sets_node = Node("Sets", parent=root)
    else:
        root = tk.Tk()
        root.title("Seal5 Viewer")

        tree = ttk.Treeview(root, columns=(1,))
        tree.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)

        scrollbar = ttk.Scrollbar(root, orient=tk.VERTICAL, command=tree.yview)
        tree.configure(yscroll=scrollbar.set)
        scrollbar.pack(side=tk.LEFT, fill=tk.Y)

        # tree.heading(0, text="Item")
        tree.heading("#0", text="Item")
        tree.heading(1, text="Value")

    # add each core to the treeview
    for set_name, set_def in sorted(model_obj.sets.items()):
        if args.text:
            set_node = Node("Set", parent=sets_node, value=set_name)
            # add constants to tree
            # consts_node = Node("Constants", parent=set_node)
            # TODO
        else:
            set_id = tree.insert("", tk.END, text=set_name)

            # add constants to tree
            consts_id = tree.insert(set_id, tk.END, text="Constants")
        for const_name, const_def in sorted(set_def.constants.items()):
            if args.text:
                _ = Node(f"{const_name}", parent=consts_id, value=const_def.value)
            else:
                tree.insert(consts_id, tk.END, text=const_name, values=(const_def.value,))

        # add memories to tree
        if args.text:
            mems_node = Node("Memories", parent=set_node)
        else:
            mems_id = tree.insert(set_id, tk.END, text="Memories")
        for mem_name, mem_def in sorted(set_def.memories.items()):
            if args.text:
                _ = Node(
                    f"{mem_name}",
                    parent=mems_node,
                    value=f"{mem_def.range.upper}:{mem_def.range.lower} ({mem_def.range.length}), {mem_def.size}",
                )
            else:
                tree.insert(
                    mems_id,
                    tk.END,
                    text=mem_name,
                    values=(f"{mem_def.range.upper}:{mem_def.range.lower} ({mem_def.range.length}), {mem_def.size}",),
                )

        # add memory aliases to tree
        if args.text:
            alias_node = Node("Memory Aliases", parent=set_node)
        else:
            alias_id = tree.insert(set_id, tk.END, text="Memory Aliases")
        for mem_name, mem_def in sorted(set_def.memory_aliases.items()):
            if args.text:
                _ = Node(
                    f"{mem_name} ({mem_def.parent.name})",
                    parent=alias_node,
                    value=f"{mem_def.range.upper}:{mem_def.range.lower} ({mem_def.range.length}), {mem_def.size}",
                )
            else:
                tree.insert(
                    alias_id,
                    tk.END,
                    text=f"{mem_name} ({mem_def.parent.name})",
                    values=(f"{mem_def.range.upper}:{mem_def.range.lower} ({mem_def.range.length}), {mem_def.size}",),
                )

        # add auxillary attributes
        # tree.insert(core_id, tk.END, text="Main Memory Object", values=(core_def.main_memory,))
        # tree.insert(core_id, tk.END, text="Main Register File Object", values=(core_def.main_reg_file,))
        # tree.insert(core_id, tk.END, text="PC Memory Object", values=(core_def.pc_memory,))

        # add functions to tree
        if args.text:
            fns_node = Node("Functions", parent=set_node)
        else:
            fns_id = tree.insert(set_id, tk.END, text="Functions")
        for fn_name, fn_def in set_def.functions.items():
            if args.text:
                fn_node = Node(fn_name, parent=fns_node, value="extern" if fn_def.extern else "")
            else:
                fn_id = tree.insert(fns_id, tk.END, text=fn_name, values=("extern" if fn_def.extern else ""))

            # add returns and throws information
            return_str = "None" if fn_def.size is None else f"{fn_def.data_type} {fn_def.size}"
            if args.text:
                _ = Node("Return", parent=fn_node, value=return_str)
                _ = Node("Throws", parent=fn_node, value=fn_def.throws)
            else:
                tree.insert(fn_id, tk.END, text="Return", values=(return_str,))
                tree.insert(fn_id, tk.END, text="Throws", values=(fn_def.throws,))

            # generate and add attributes
            if args.text:
                attrs_node = Node("Attributes", parent=fn_node)
            else:
                attrs_id = tree.insert(fn_id, tk.END, text="Attributes")

            for attr, ops in fn_def.attributes.items():
                if args.text:
                    attr_node = Node(attr, parent=attrs_node)
                else:
                    attr_id = tree.insert(attrs_id, tk.END, text=attr)
                for op in ops:
                    if args.text:
                        context = TextTreeGenContext(parent=attr_node)
                    else:
                        context = TkTreeGenContext(tree, parent=attr_id)
                    if args.operation:
                        op.generate(context)

            # generate and add parameters
            if args.text:
                params_node = Node("Parameters", parent=fn_node)
            else:
                params_id = tree.insert(fn_id, tk.END, text="Parameters")

            for param_name, param_def in fn_def.args.items():
                if args.text:
                    _ = Node(param_name, parent=params_node, value=f"{param_def.data_type} {param_def.size}")
                else:
                    tree.insert(params_id, tk.END, text=param_name, values=(f"{param_def.data_type} {param_def.size}",))

            # generate and add function behavior
            if args.text:
                context = TextTreeGenContext(parent=fn_node)
            else:
                context = TkTreeGenContext(tree, parent=fn_id)
            if args.operation:
                fn_def.operation.generate(context)

        # group instructions by size
        instrs_by_size = defaultdict(dict)

        for k, v in set_def.instructions.items():
            instrs_by_size[v.size][k] = v

        # sort instructions by encoding
        for k, v in instrs_by_size.items():
            instrs_by_size[k] = dict(sorted(v.items(), key=sort_instruction, reverse=True))

        if args.text:
            instrs_top_node = Node("Instructions", parent=set_node)
        else:
            instrs_top_id = tree.insert(set_id, tk.END, text="Instructions")

        # generate instruction size groups
        for size, instrs in sorted(instrs_by_size.items()):
            if args.text:
                instrs_node = Node(f"Width {size}", parent=instrs_top_node)
            else:
                instrs_id = tree.insert(instrs_top_id, tk.END, text=f"Width {size}")

            # generate instructions
            for (code, mask), instr_def in instrs.items():
                opcode_str = "{code:0{width}x}:{mask:0{width}x}".format(
                    code=code, mask=mask, width=int(instr_def.size / 4)
                )

                if args.text:
                    instr_node = Node(f"{instr_def.ext_name} : {instr_def.name}", parent=instrs_node, value=opcode_str)
                else:
                    instr_id = tree.insert(
                        instrs_id,
                        tk.END,
                        text=f"{instr_def.ext_name} : {instr_def.name}",
                        values=(opcode_str,),
                        tags=("mono",),
                    )

                # generate encoding
                enc_str = []
                for enc in instr_def.encoding:
                    if isinstance(enc, arch.BitVal):
                        enc_str.append(f"{enc.value:0{enc.length}b}")
                    elif isinstance(enc, arch.BitField):
                        enc_str.append(f"{enc.name}[{enc.range.upper}:{enc.range.lower}]")

                if args.text:
                    _ = Node("Encoding", parent=instr_node, value=" ".join(enc_str))
                    _ = Node("Assembly", parent=instr_node, value=instr_def.assembly)
                    _ = Node("Throws", parent=instr_node, value=instr_def.throws)
                    attrs_node = Node("Attributes", parent=instr_node)
                else:
                    tree.insert(instr_id, tk.END, text="Encoding", values=(" ".join(enc_str),))
                    tree.insert(instr_id, tk.END, text="Assembly", values=(instr_def.assembly,))
                    tree.insert(instr_id, tk.END, text="Throws", values=(instr_def.throws,))
                    attrs_id = tree.insert(instr_id, tk.END, text="Attributes")

                # generate attributes
                for attr, ops in instr_def.attributes.items():
                    if args.text:
                        attr_node = Node(attr.name, parent=attrs_node)
                    else:
                        attr_id = tree.insert(attrs_id, tk.END, text=attr.name)
                    for op in ops:
                        if args.text:
                            context = TextTreeGenContext(parent=attr_node)
                        else:
                            context = TkTreeGenContext(tree, parent=attr_id)
                        if args.operation:
                            op.generate(context)

                # _ = Node("Behavior", parent=instr_node, value=instr_def.operation.statements)
                # generate behavior
                if args.text:
                    context = TextTreeGenContext(parent=instr_node)
                else:
                    context = TkTreeGenContext(tree, parent=instr_id)
                if args.operation:
                    instr_def.operation.generate(context)

    if args.text:
        print("============================")
        # text = ""
        for pre, _, node in RenderTree(root):
            suffix = ""
            if hasattr(node, "value"):
                if node.value is not None:
                    suffix = f" [{node.value}]"
            print(f"{pre}{node.name}{suffix}")
        print("============================")

    else:
        # tree.tag_configure("mono", font=font.nametofont("TkFixedFont"))

        root.mainloop()


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

from m2isar import M2NameError, M2SyntaxError, M2TypeError, flatten
from m2isar.metamodel import arch, behav, intrinsics
from m2isar.metamodel.utils import StaticType
from .parser_gen import CoreDSL2Parser, CoreDSL2Visitor
from .utils import BOOLCONST, RADIX, SHORTHANDS, SIGNEDNESS

# import seal5.model as seal5_model

from seal5.logging import Logger

logger = Logger("frontends.behav_builder")


class BehaviorModelBuilder(CoreDSL2Visitor):
    """ANTLR visitor to build an M2-ISA-R behavioral model of a function or instruction
    of a CoreDSL 2 specification.
    """

    def __init__(
        self,
        constants: "dict[str, arch.Constant]",
        memories: "dict[str, arch.Memory]",
        memory_aliases: "dict[str, arch.Memory]",
        fields: "dict[str, arch.BitFieldDescr]",
        functions: "dict[str, arch.Function]",
        warned_fns: "set[str]",
    ):
        super().__init__()

        self._constants = constants
        self._memories = memories
        self._memory_aliases = memory_aliases
        self._fields = fields
        self._scalars = {}
        self._functions = functions
        self.warned_fns = warned_fns if warned_fns is not None else set()

    def visitChildren(self, node):
        """Helper method to return flatter results on tree visits."""

        ret = super().visitChildren(node)
        if isinstance(ret, list) and len(ret) == 1:
            return ret[0]
        return ret

    def aggregateResult(self, aggregate, nextResult):
        """Aggregate results from multiple children into a list."""

        ret = aggregate
        if nextResult is not None:
            if ret is None:
                ret = [nextResult]
            else:
                ret += [nextResult]
        return ret

    def visitProcedure_call(self, ctx: CoreDSL2Parser.Procedure_callContext):
        """Generate a procedure (method call without return value) call."""

        # extract name and reference to procedure object to be called
        name = ctx.ref.text
        ref = self._functions.get(name, None)

        # error out if method is unknown
        if ref is None:
            raise M2NameError(f"procedure {name} is not defined")

        # generate method arguments
        args = [self.visit(obj) for obj in ctx.args] if ctx.args else []

        return behav.ProcedureCall(ref, args)

    def visitMethod_call(self, ctx: "CoreDSL2Parser.Method_callContext"):
        """Generate a function (method call with return value) call."""

        # extract name and reference to function object to be called
        name = ctx.ref.text
        ref = self._functions.get(name, None)

        # error out if method is unknown
        if ref is None:
            raise M2NameError(f'function "{name}" is not defined')

        if arch.FunctionAttribute.ETISS_TRAP_ENTRY_FN in ref.attributes:
            raise M2SyntaxError(f'exception entry function "{name}" must be called as procedure')

        # generate method arguments
        args = [self.visit(obj) for obj in ctx.args] if ctx.args else []

        return behav.FunctionCall(ref, args)

    def visitBlock(self, ctx: CoreDSL2Parser.BlockContext):
        """Generate a block of statements, return a list."""

        items = [self.visit(obj) for obj in ctx.items]
        items = list(flatten(items))
        return behav.Block(items)

    def visitDeclaration(self, ctx: CoreDSL2Parser.DeclarationContext):
        """Generate a declaration statement. Can be multiple declarations of
        the same type at once. Each declaration can have an initial value.
        """

        # extract variable qualifiers, currently unused
        # storage = [self.visit(obj) for obj in ctx.storage]
        # qualifiers = [self.visit(obj) for obj in ctx.qualifiers]
        # attributes = [self.visit(obj) for obj in ctx.attributes]

        type_ = self.visit(ctx.type_)

        decls: "list[CoreDSL2Parser.DeclaratorContext]" = ctx.declarations

        ret_decls = []

        # iterate over all contained declarations
        for decl in decls:
            name = decl.name.text

            # instantiate a scalar and its definition
            s = arch.Scalar(
                name, None, StaticType.NONE, type_.width, arch.DataType.S if type_.signed else arch.DataType.U
            )
            self._scalars[name] = s
            sd = behav.ScalarDefinition(s)

            # if initializer is present, generate an assignment to apply
            # initialization to the scalar
            if decl.init:
                init = self.visit(decl.init)
            else:
                init = behav.IntLiteral(0)

            a = behav.Assignment(sd, init)
            ret_decls.append(a)

        return ret_decls

    def visitBreak_statement(self, ctx: CoreDSL2Parser.Break_statementContext):
        return behav.Break()

    def visitReturn_statement(self, ctx: CoreDSL2Parser.Return_statementContext):
        """Generate a return statement."""

        expr = self.visit(ctx.expr) if ctx.expr else None
        return behav.Return(expr)

    def visitWhile_statement(self, ctx: CoreDSL2Parser.While_statementContext):
        """Generate a while loop."""

        stmt = self.visit(ctx.stmt) if ctx.stmt else None
        cond = self.visit(ctx.cond)

        if not isinstance(stmt, list):
            stmt = [stmt]

        return behav.Loop(cond, stmt, False)

    def visitDo_statement(self, ctx: CoreDSL2Parser.Do_statementContext):
        """Generate a do .. while loop."""

        stmt = self.visit(ctx.stmt) if ctx.stmt else None
        cond = self.visit(ctx.cond)

        if not isinstance(stmt, list):
            stmt = [stmt]

        return behav.Loop(cond, stmt, True)

    def visitFor_statement(self, ctx: CoreDSL2Parser.For_statementContext):
        """Generate a for loop. Currently hacky, untested and mostly broken."""

        start_decl, start_expr, end_expr, loop_exprs = self.visit(ctx.cond)
        stmt = self.visit(ctx.stmt) if ctx.stmt else None

        if not isinstance(stmt, list):
            stmt = [stmt]

        ret = []

        if start_decl is not None:
            ret.append(start_decl)
        if start_expr is not None:
            ret.append(start_expr)

        if loop_exprs:
            stmt.extend(loop_exprs)

        ret.append(behav.Loop(end_expr, stmt, False))

        return ret

    def visitFor_condition(self, ctx: CoreDSL2Parser.For_conditionContext):
        """Generate the condition of a for loop."""

        start_decl = self.visit(ctx.start_decl) if ctx.start_decl else None
        start_expr = self.visit(ctx.start_expr) if ctx.start_expr else None
        end_expr = self.visit(ctx.end_expr) if ctx.end_expr else None
        loop_exprs = [self.visit(obj) for obj in ctx.loop_exprs] if ctx.loop_exprs else None

        return start_decl, start_expr, end_expr, loop_exprs

    def visitIf_statement(self, ctx: CoreDSL2Parser.If_statementContext):
        """Generate an if statement. Packs all if, else if and else branches
        into one object.
        """

        conds = [self.visit(x) for x in ctx.cond]
        stmts = [self.visit(x) for x in ctx.stmt]

        stmts = [x if not isinstance(x, list) else None for x in stmts]

        if None in stmts:
            raise Exception("meep")

        return behav.Conditional(conds, stmts)

    def visitConditional_expression(self, ctx: CoreDSL2Parser.Conditional_expressionContext):
        """Generate a ternary expression."""

        cond = self.visit(ctx.cond)
        then_expr = self.visit(ctx.then_expr)
        else_expr = self.visit(ctx.else_expr)

        return behav.Ternary(cond, then_expr, else_expr)

    def visitBinary_expression(self, ctx: CoreDSL2Parser.Binary_expressionContext):
        """Generate a binary expression."""

        left = self.visit(ctx.left)
        op = behav.Operator(ctx.bop.text)
        right = self.visit(ctx.right)

        return behav.BinaryOperation(left, op, right)

    def visitPreinc_expression(self, ctx: CoreDSL2Parser.Preinc_expressionContext):
        """Generate a pre-increment expression. Not yet supported, throws
        :exc:`NotImplementedError`."""

        raise NotImplementedError("pre-increment expressions are not supported yet")

    def visitPostinc_expression(self, ctx: CoreDSL2Parser.Preinc_expressionContext):
        """Generate a post-increment expression. Not yet supported, throws
        :exc:`NotImplementedError`."""

        raise NotImplementedError("post-increment expressions are not supported yet")

    def visitPrefix_expression(self, ctx: CoreDSL2Parser.Prefix_expressionContext):
        """Generate an unary expression."""

        op = behav.Operator(ctx.prefix.text)
        right = self.visit(ctx.right)

        return behav.UnaryOperation(op, right)

    def visitParens_expression(self, ctx: CoreDSL2Parser.Parens_expressionContext):
        """Generate a parenthesized expression."""

        expr = self.visit(ctx.expr)
        return behav.Group(expr)

    def visitSlice_expression(self, ctx: CoreDSL2Parser.Slice_expressionContext):
        """Generate a slice expression. Depending on context, this is translated
        to either an actual :class:`m2isar.metamodel.behav.SliceOperation`or
        an :class:`m2isar.metamodel.behav.IndexedReference` if a :class:`m2isar.metamodel.arch.Memory
        object is to be sliced.
        """

        expr = self.visit(ctx.expr)

        left = self.visit(ctx.left)
        right = self.visit(ctx.right) if ctx.right else left

        if (
            isinstance(expr, behav.NamedReference)
            and isinstance(expr.reference, arch.Memory)
            and (expr.reference.data_range is None or expr.reference.data_range.length > 1)
        ):
            return behav.IndexedReference(expr.reference, left, right)
        else:
            return behav.SliceOperation(expr, left, right)

    def visitConcat_expression(self, ctx: CoreDSL2Parser.Concat_expressionContext):
        """Generate a concatenation expression."""

        left = self.visit(ctx.left)
        right = self.visit(ctx.right)

        return behav.ConcatOperation(left, right)

    def visitAssignment_expression(self, ctx: CoreDSL2Parser.Assignment_expressionContext):
        """Generate an assignment. If a combined arithmetic-assignment is present,
        generate an additional binary operation and use it as the RHS.
        """

        op = ctx.bop.text
        left = self.visit(ctx.left)
        right = self.visit(ctx.right)

        if op != "=":
            op2 = behav.Operator(op[:-1])
            right = behav.BinaryOperation(left, op2, right)

        return behav.Assignment(left, right)

    def visitReference_expression(self, ctx: CoreDSL2Parser.Reference_expressionContext):
        """Generate a simple reference."""

        name = ctx.ref.text

        var = (
            self._scalars.get(name)
            or self._fields.get(name)
            or self._constants.get(name)
            or self._memory_aliases.get(name)
            or self._memories.get(name)
            or intrinsics.get(name)
        )

        if var is None:
            raise M2NameError(f'Named reference "{name}" does not exist!')

        return behav.NamedReference(var)

    def visitInteger_constant(self, ctx: CoreDSL2Parser.Integer_constantContext):
        """Generate an integer literal."""

        text: str = ctx.value.text.lower()

        tick_pos = text.find("'")

        if tick_pos != -1:
            width = int(text[:tick_pos])
            radix = text[tick_pos + 1]
            value = int(text[tick_pos + 2 :], RADIX[radix])

        else:
            value = int(text, 0)
            width = value.bit_length()

        return behav.IntLiteral(value, width)

    def visitCharacter_constant(self, ctx: CoreDSL2Parser.Character_constantContext):
        """Generate a character literal. Converts directly to uint8."""

        text: str = ctx.value.text

        value = min(ord(text.replace("'", "")), 255)

        return behav.IntLiteral(value, 8)

    # def visitString_literal(self, ctx:CoreDSL2Parser.String_literalContext):

    #     print("visitString_literal", self, ctx, dir(self), dir(ctx))
    #     text: str = ctx.value.text
    #     # text: str = ctx.text

    #     return behav.IntLiteral(0, 8)

    def visitString_constant(self, ctx: CoreDSL2Parser.String_constantContext):
        # print("visitString_constant", self, ctx, dir(self), dir(ctx))
        text: str = ctx.value.text
        assert len(text) >= 2
        assert text[0] == '"' and text[-1] == '"'
        text = text[1:-1]
        # text: str = ctx.text

        return behav.StringLiteral(text)
        # return seal5_model.StringLiteral(text)

    def visitBool_constant(self, ctx: CoreDSL2Parser.Bool_constantContext):
        """Generate a boolean literal. Converts directly to uint1."""

        text: str = ctx.value.text

        return behav.IntLiteral(BOOLCONST[text], 1)

    def visitCast_expression(self, ctx: CoreDSL2Parser.Cast_expressionContext):
        """Generate a type cast."""

        expr = self.visit(ctx.right)
        if ctx.type_:
            type_ = self.visit(ctx.type_)
            sign = arch.DataType.S if type_.signed else arch.DataType.U
            size = type_.width

        if ctx.sign:
            sign = self.visit(ctx.sign)
            sign = arch.DataType.S if sign else arch.DataType.U
            size = None

        return behav.TypeConv(sign, size, expr)

    def visitType_specifier(self, ctx: CoreDSL2Parser.Type_specifierContext):
        """Generate a generic type specifier."""

        type_ = self.visit(ctx.type_)
        if ctx.ptr:
            type_.ptr = ctx.ptr.text
        return type_

    def visitInteger_type(self, ctx: CoreDSL2Parser.Integer_typeContext):
        """Generate an integer type specifier."""

        signed = True
        width = None

        if ctx.signed is not None:
            signed = self.visit(ctx.signed)

        if ctx.size is not None:
            width = self.visit(ctx.size)

        if ctx.shorthand is not None:
            width = self.visit(ctx.shorthand)

        if isinstance(width, behav.BaseNode):
            width = width.generate(None)
        else:
            raise M2TypeError("width has wrong type")

        return arch.IntegerType(width, signed, None)

    def visitVoid_type(self, ctx: CoreDSL2Parser.Void_typeContext):
        """Generate a void type specifier."""

        return arch.VoidType(None)

    def visitBool_type(self, ctx: CoreDSL2Parser.Bool_typeContext):
        """Generate a bool type specifier. Aliases to unsigned<1>."""

        return arch.IntegerType(1, False, None)

    def visitInteger_signedness(self, ctx: CoreDSL2Parser.Integer_signednessContext):
        """Generate integer signedness."""

        return SIGNEDNESS[ctx.children[0].symbol.text]

    def visitInteger_shorthand(self, ctx: CoreDSL2Parser.Integer_shorthandContext):
        """Lookup a shorthand type specifier."""

        return behav.IntLiteral(SHORTHANDS[ctx.children[0].symbol.text])



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Annotate instruction which are unsupported with SKIP_PATTERN_GEN attribute."""

import sys
import argparse
import logging
import pathlib

import pandas as pd

from m2isar.metamodel import arch, behav
from seal5.model import Seal5InstrAttribute, Seal5OperandAttribute
from seal5.model_utils import load_model, dump_model


from seal5.logging import Logger

logger = Logger("transform.check_pattern_support")


def get_parser():
    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--metrics", default=None, help="Output metrics to file")
    parser.add_argument("--compat", action="store_true")
    return parser


def run(args):
    # initialize logging
    logger.setLevel(getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)

    out_path = (top_level.parent / top_level.stem) if args.output is None else args.output

    model_obj = load_model(top_level, compat=args.compat)

    metrics = {
        "n_sets": 0,
        "n_instructions": 0,
        "n_skipped": 0,
        "n_failed": 0,
        "n_success": 0,
        "skipped_instructions": [],
        "failed_instructions": [],
        "success_instructions": [],
    }
    for _, set_def in model_obj.sets.items():
        metrics["n_sets"] += 1
        logger.debug("checking set %s", set_def.name)
        for _, instr_def in set_def.instructions.items():
            metrics["n_instructions"] += 1
            logger.debug("checking instr %s", instr_def.name)
            try:
                attributes = instr_def.attributes

                def check_noop(operation):
                    if len(operation.statements) == 0:
                        return True
                    if len(operation.statements) == 1:
                        if isinstance(operation.statements[0], behav.Block):
                            return len(operation.statements[0].statements) == 0
                    return False

                is_noop = check_noop(instr_def.operation)
                num_ins = len(
                    [
                        op
                        for op in instr_def.operands.values()
                        if Seal5OperandAttribute.IN in op.attributes or Seal5OperandAttribute.INOUT in op.attributes
                    ]
                )
                num_outs = len(
                    [
                        op
                        for op in instr_def.operands.values()
                        if Seal5OperandAttribute.OUT in op.attributes or Seal5OperandAttribute.INOUT in op.attributes
                    ]
                )
                may_load = Seal5InstrAttribute.MAY_LOAD in attributes
                may_store = Seal5InstrAttribute.MAY_STORE in attributes
                is_rvc = instr_def.size != 32
                is_branch = arch.InstrAttribute.COND in attributes or arch.InstrAttribute.NO_CONT in attributes
                has_loop = Seal5InstrAttribute.HAS_LOOP in attributes
                # TODO: has_static_loop
                has_call = Seal5InstrAttribute.HAS_CALL in attributes
                uses_custom_reg = len(attributes.get(Seal5InstrAttribute.USES, []))
                defs_custom_reg = len(attributes.get(Seal5InstrAttribute.DEFS, []))
                # TODO: check if PC is being read/written -> not supported
                # uses_pc = ?
                skip_pattern_gen = (
                    is_noop
                    or is_rvc
                    or (num_ins == 0)
                    or (num_outs != 1)
                    or may_load
                    or may_store
                    or is_branch
                    or has_loop
                    or has_call
                    or uses_custom_reg
                    or defs_custom_reg
                )
                if skip_pattern_gen:
                    attributes[Seal5InstrAttribute.SKIP_PATTERN_GEN] = []
                # TODO: move to different/new pass (add_instr_attributes) and
                # move detections to Seal5Instr class in metamodel
                attributes[Seal5InstrAttribute.LLVM_INSTR] = behav.StringLiteral(instr_def.name)
                instr_def.attributes = attributes
                metrics["n_success"] += 1
                metrics["success_instructions"].append(instr_def.name)
            except Exception as ex:
                logger.exception(ex)
                metrics["n_failed"] += 1
                metrics["failed_instructions"].append(instr_def.name)

    dump_model(model_obj, out_path, compat=args.compat)
    if args.metrics:
        metrics_file = args.metrics
        metrics_df = pd.DataFrame({key: [val] for key, val in metrics.items()})
        metrics_df.to_csv(metrics_file, index=False)


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    run(args)


if __name__ == "__main__":
    main(sys.argv[1:])



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Create LLVM patches for new field types."""

import re
import argparse
import logging
import pathlib

from seal5.index import write_index_yaml, NamedPatch

from seal5.settings import Seal5Settings

from seal5.logging import Logger

logger = Logger("backends.riscv_field_types")

SEAL5_RISCV_FIELDS_SUPPORT = """class Seal5RISCVUImmOp<int bitsNum> : RISCVOp {
  let ParserMatchClass = UImmAsmOperand<bitsNum>;
  let DecoderMethod = "decodeUImmOperand<" # bitsNum # ">";
  let OperandType = "SEAL5_OPERAND_UIMM" # bitsNum;
}
class Seal5RISCVUImmLeafOp<int bitsNum> :
  Seal5RISCVUImmOp<bitsNum>, ImmLeaf<XLenVT, "return isUInt<" # bitsNum # ">(Imm);">;
class Seal5RISCVSImmOp<int bitsNum> : RISCVOp {
  let ParserMatchClass = SImmAsmOperand<bitsNum>;
  let EncoderMethod = "getImmOpValue";
  let DecoderMethod = "decodeSImmOperand<" # bitsNum # ">";
  let OperandType = "SEAL5_OPERAND_SIMM" # bitsNum;
}
class Seal5RISCVSImmLeafOp<int bitsNum> :
  Seal5RISCVSImmOp<bitsNum>, ImmLeaf<XLenVT, "return isInt<" # bitsNum # ">(Imm);">;
"""


def gen_riscv_field_types_str(field_types):
    # print("gen_riscv_field_types_str", field_types)
    riscv_field_types_contents = [SEAL5_RISCV_FIELDS_SUPPORT] if len(field_types) > 0 else []
    riscv_operands_asm_contents = []
    riscv_operands_enum_contents = []

    for field_type in field_types:
        # print("field_type", field_type)
        matches = re.compile(r"([us])imm(\d+|log2xlen)([_a-zA-Z].*)?").match(field_type)
        assert matches is not None, f"Field type not supported: {field_type}"
        # print("matches", matches)
        groups = list(matches.groups())
        sign_letter, imm_size, suffix = groups
        # print("sign_letter", sign_letter)
        # print("imm_size", imm_size)
        # print("suffix", suffix)
        assert sign_letter in ["u", "s"]
        assert imm_size.isdigit()
        imm_size = int(imm_size)
        assert suffix is None  # TODO: handle suffixes
        # TODO: leaf or not?
        is_leaf = True
        if sign_letter == "u":
            cls = "Seal5RISCVUImmLeafOp" if is_leaf else "Seal5RISCVUImmOp"
            temp = f"def {field_type} : {cls}<{imm_size}>;"
        elif sign_letter == "s":
            cls = "Seal5RISCVSImmLeafOp" if is_leaf else "Seal5RISCVSImmOp"
            # Warning: RISCVSImmOp not tested yet
            temp = f"""def {field_type} : {cls}<{imm_size}> {{
  let MCOperandPredicate = [{{
    int64_t Imm;
    if (MCOp.evaluateAsConstantImm(Imm))
      return isInt<{imm_size}>(Imm);
    return MCOp.isBareSymbolRef();
  }}];
}}"""
        else:
            assert False  # Should not be reached
        riscv_field_types_contents.append(temp)
        sign_letter_upper = sign_letter.upper()
        if sign_letter_upper == "U":
            temp = (
                f"bool is{sign_letter_upper}Imm{imm_size}() const {{ return Is{sign_letter_upper}Imm<{imm_size}>(); }}"
            )
        elif sign_letter_upper == "S":
            # TODO: introduce `template <unsigned N> bool IsSImm() const {`
            # to make this more clean
            temp = f"""bool is{sign_letter_upper}Imm{imm_size}() const {{
      if (!isImm())
        return false;
      RISCVMCExpr::VariantKind VK = RISCVMCExpr::VK_RISCV_None;
      int64_t Imm;
      bool IsConstantImm = evaluateConstantImm(getImm(), Imm, VK);
      return IsConstantImm && isInt<{imm_size}>(fixImmediateForRV32(Imm, isRV64Imm())) &&
             VK == RISCVMCExpr::VK_RISCV_None;
    }}"""
        else:
            assert False  # Should not be reached
        riscv_operands_asm_contents.append(temp)
        field_type_upper = field_type.upper()
        temp = f"  SEAL5_OPERAND_{field_type_upper},"
        riscv_operands_enum_contents.append(temp)

    riscv_field_types_content = "\n".join(riscv_field_types_contents)
    riscv_operands_asm_content = "\n".join(riscv_operands_asm_contents)
    riscv_operands_enum_content = "\n".join(riscv_operands_enum_contents)
    return riscv_field_types_content, riscv_operands_asm_content, riscv_operands_enum_content


def main():
    """Main app entrypoint."""

    # read command line args
    parser = argparse.ArgumentParser()
    # parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")  # TODO: drop?
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--metrics", default=None, help="Output metrics to file")
    parser.add_argument("--index", default=None, help="Output index to file")
    parser.add_argument("--ext", type=str, default="td", help="Default file extension (if using --splitted)")
    parser.add_argument("--compat", action="store_true")
    parser.add_argument("--yaml", type=str, default=None)
    args = parser.parse_args()

    # initialize logging
    logger.setLevel(getattr(logging, args.log.upper()))

    # resolve model paths
    # top_level = pathlib.Path(args.top_level)
    assert args.yaml is not None
    assert pathlib.Path(args.yaml).is_file()
    settings = Seal5Settings.from_yaml_file(args.yaml)

    # print("settings", settings)
    # print("settings.models", settings.models)
    # print("settings.models['Example']", settings.models["Example"])
    # print("settings.models['Example'].extensions", settings.models["Example"].extensions)
    # print("settings.models['Example'].extensions['XExample']", settings.models["Example"].extensions["XExample"])
    # print(
    #     "settings.models['Example'].extensions['XExample'].required_imm_types",
    #     settings.models["Example"].extensions["XExample"].required_imm_types,
    # )
    # print("settings.models", settings.models)
    all_required_imm_types = set(
        sum(
            [
                ext_settings.required_imm_types
                for model_name, model_settings in settings.models.items()
                for ext_name, ext_settings in model_settings.extensions.items()
                if ext_settings.required_imm_types is not None
            ],
            [],
        )
    )
    # print("all_required_imm_types", all_required_imm_types)
    supported_imm_types = set(settings.llvm.state.supported_imm_types)
    # print("supported_imm_types", supported_imm_types)
    missing_imm_types = all_required_imm_types - supported_imm_types
    # print("missing_imm_types", missing_imm_types)

    metrics = {
        "n_imm": 0,
        "n_skipped": 0,
        "n_failed": 0,
        "n_success": 0,
    }
    # preprocess model
    # print("model", model)
    # settings = model.get("settings", None)
    artifacts = {}
    artifacts[None] = []  # used for global artifacts
    # TODO: error handling?
    if len(missing_imm_types) > 0:
        field_types_content, riscv_operands_asm_content, riscv_operands_enum_content = gen_riscv_field_types_str(
            missing_imm_types
        )
    else:
        field_types_content, riscv_operands_asm_content, riscv_operands_enum_content = "", "", ""
    metrics["n_imm"] = len(missing_imm_types)
    metrics["n_success"] = len(missing_imm_types)
    # print("field_types_content", field_types_content)
    # print("riscv_operands_asm_content", riscv_operands_asm_content)
    # print("riscv_operands_enum_content", riscv_operands_enum_content)
    if len(field_types_content) > 0:
        field_types_patch = NamedPatch(
            "llvm/lib/Target/RISCV/RISCVInstrInfo.td",
            key="field_types",  # TODO: rename: riscv_field_types
            content=field_types_content,
        )
        artifacts[None].append(field_types_patch)
    if len(riscv_operands_asm_content) > 0:
        riscv_operands_asm_patch = NamedPatch(
            "llvm/lib/Target/RISCV/AsmParser/RISCVAsmParser.cpp",
            key="riscv_operands",  # TODO: rename: riscv_operands_asm
            content=riscv_operands_asm_content,
        )
        artifacts[None].append(riscv_operands_asm_patch)
    if len(riscv_operands_enum_content) > 0:
        riscv_operands_enum_patch = NamedPatch(
            "llvm/lib/Target/RISCV/MCTargetDesc/RISCVBaseInfo.h",
            key="riscv_operands",  # TODO: rename: riscv_operands_enum
            content=riscv_operands_enum_content,
        )
        artifacts[None].append(riscv_operands_enum_patch)
    # input("!!!")
    if args.metrics:
        metrics_file = args.metrics
        with open(metrics_file, "w", encoding="utf-8") as f:
            f.write(",".join(metrics.keys()))
            f.write("\n")
            f.write(",".join(map(str, metrics.values())))
            f.write("\n")
    if args.index:
        if sum(map(len, artifacts.values())) > 0:
            global_artifacts = artifacts.get(None, [])
            set_artifacts = {key: value for key, value in artifacts.items() if key is not None}
            index_file = args.index
            write_index_yaml(index_file, global_artifacts, set_artifacts, content=True)
        else:
            logger.warning("No patches generated. No index file will be written.")


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""A transformation module for simplifying M2-ISA-R behavior expressions. The following
simplifications are done:

* Resolvable :class:`m2isar.metamodel.arch.Constant` s are replaced by
  `m2isar.metamodel.arch.IntLiteral` s representing their value
* Fully resolvable arithmetic operations are carried out and their results
  represented as a matching :class:`m2isar.metamodel.arch.IntLiteral`
* Conditions and loops with fully resolvable conditions are either discarded entirely
  or transformed into code blocks without any conditions
* Ternaries with fully resolvable conditions are transformed into only the matching part
* Type conversions of :class:`m2isar.metamodel.arch.IntLiteral` s apply the desired
  type directly to the :class:`IntLiteral` and discard the type conversion
"""
from m2isar.metamodel import arch, behav

from seal5.logging import Logger

logger = Logger("transform." + __name__)


# pylint: disable=unused-argument


def operation(self: behav.Operation, context):
    statements = []
    for stmt in self.statements:
        try:
            temp = stmt.generate(context)
            if isinstance(temp, list):
                statements.extend(temp)
            else:
                statements.append(temp)
        except (NotImplementedError, ValueError):
            print(f"cant simplify {stmt}")

    self.statements = statements
    return self


def binary_operation(self: behav.BinaryOperation, context):
    self.left = self.left.generate(context)
    self.right = self.right.generate(context)

    return self


def slice_operation(self: behav.SliceOperation, context):
    # print("slice_operation")
    self.expr = self.expr.generate(context)

    if self.expr.inferred_type is None:
        logger.warning("Slice Operation needs inferred type. Skipping...")
        return self
    assert isinstance(self.expr.inferred_type, arch.IntegerType)
    source_width = self.expr.inferred_type.width
    # print("source_width", source_width)
    self.left = self.left.generate(context)
    if not isinstance(self.left, behav.IntLiteral):
        return self
    self.right = self.right.generate(context)
    if not isinstance(self.right, behav.IntLiteral):
        return self
    assert self.left.value >= self.right.value
    target_width = self.left.value - self.right.value + 1
    # print("target_width", target_width)
    assert target_width <= source_width

    # input("@")
    if source_width == target_width:
        return behav.Group(self.expr)

    # if self.right.value == 0:
    #     return behav.TypeConv(arch.DataType.S if self.expr.inferred_type.signed
    #        else arch.DataType.U, target_width, self.expr)
    return self


def concat_operation(self: behav.ConcatOperation, context):
    self.left = self.left.generate(context)
    self.right = self.right.generate(context)

    return self


def number_literal(self: behav.IntLiteral, context):
    return self


def int_literal(self: behav.IntLiteral, context):
    return self


def scalar_definition(self: behav.ScalarDefinition, context):
    return self


def assignment(self: behav.Assignment, context):
    self.target = self.target.generate(context)
    self.expr = self.expr.generate(context)

    # if isinstance(self.expr, behav.IntLiteral) and isinstance(self.target, behav.ScalarDefinition):
    #       self.target.scalar.value = self.expr.value

    return self


def conditional(self: behav.Conditional, context):
    self.conds = [x.generate(context) for x in self.conds]
    # self.stmts = [[y.generate(context) for y in x] for x in self.stmts]
    self.stmts = [x.generate(context) for x in self.stmts]

    return self


def loop(self: behav.Loop, context):
    self.cond = self.cond.generate(context)
    self.stmts = [x.generate(context) for x in self.stmts]

    return self


def ternary(self: behav.Ternary, context):
    self.cond = self.cond.generate(context)
    self.then_expr = self.then_expr.generate(context)
    self.else_expr = self.else_expr.generate(context)

    return self


def return_(self: behav.Return, context):
    if self.expr is not None:
        self.expr = self.expr.generate(context)

    return self


def unary_operation(self: behav.UnaryOperation, context):
    self.right = self.right.generate(context)

    return self


def named_reference(self: behav.NamedReference, context):
    return self


def indexed_reference(self: behav.IndexedReference, context):
    self.index = self.index.generate(context)

    return self


def type_conv(self: behav.TypeConv, context):
    self.expr = self.expr.generate(context)

    return self


def callable_(self: behav.Callable, context):
    self.args = [stmt.generate(context) for stmt in self.args]

    return self


def group(self: behav.Group, context):
    self.expr = self.expr.generate(context)

    if isinstance(self.expr, behav.IntLiteral):
        return self.expr

    return self


def break_(self: behav.Break, context):
    return self



#
# Copyright (c) 2023 TUM Department of Electrical and Computer Engineering.
#
# This file is part of Seal5.
# See https://github.com/tum-ei-eda/seal5.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Command line subcommand for building Seal5 LLVM."""

from seal5.flow import Seal5Flow
from seal5.logging import Logger


logger = Logger("cli")


def add_build_options(parser):
    """Setup parser for build argument group."""
    build_parser = parser.add_argument_group("build options")
    build_parser.add_argument(
        "--target",
        "-t",
        default="all",
        help="All things that should be build in the make process",
    )
    build_parser.add_argument(
        "--config",
        default=None,
        help="Choose build Config in Settings.yml",
    )
    build_parser.add_argument(
        "--ccache",
        action="store_true",
        help="Enable CCache to speedup compilation.",
    )


def get_parser(subparsers):
    """ "Define and return a subparser for the build subcommand."""
    parser = subparsers.add_parser("build", description="Build Seal5.")
    parser.set_defaults(func=handle)
    add_build_options(parser)
    return parser


def handle(args):
    """Callback function which will be called to process the build subcommand"""
    seal5_flow = Seal5Flow(args.dir, name=args.name)
    seal5_flow.build(
        config=args.config,
        target=args.target,
        verbose=args.verbose,
        enable_ccache=args.ccache,
    )



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Clean M2-ISA-R/Seal5 metamodel to .core_desc file."""

import argparse
import logging
import pathlib

import pandas as pd
from mako.template import Template

from m2isar.metamodel import arch

from seal5.index import NamedPatch, write_index_yaml
from seal5.settings import ExtensionsSettings, LLVMVersion
from seal5.model_utils import load_model

from seal5.logging import Logger

logger = Logger("backends.riscv_isa_info")


MAKO_TEMPLATE = '    {"${arch}", RISCVExtensionVersion{${version_major}, ${version_minor}}},'
MAKO_TEMPLATE_LLVM18 = '    {"${arch}", {${version_major}, ${version_minor}}},'


def gen_riscv_isa_info_str(name: str, ext_settings: ExtensionsSettings, llvm_version: LLVMVersion):
    # print("name", name)
    # print("ext_settings", ext_settings)
    arch_ = ext_settings.get_arch(name=name)
    version = ext_settings.get_version()
    if not isinstance(version, str):
        assert isinstance(version, (int, float))
        version = str(float(version))
    version_major, version_minor = str(version).split(".", 1)

    llvm_major = llvm_version.major
    template = MAKO_TEMPLATE_LLVM18 if llvm_major >= 18 else MAKO_TEMPLATE
    content_template = Template(template)
    content_text = content_template.render(arch=arch_, version_major=version_major, version_minor=version_minor)
    # content_text = content_text.rstrip("\n")
    return arch, (content_text)


def main():
    """Main app entrypoint."""

    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--splitted", action="store_true", help="Split per set")
    parser.add_argument("--formats", action="store_true", help="Also generate instruction formats")
    parser.add_argument("--metrics", default=None, help="Output metrics to file")
    parser.add_argument("--index", default=None, help="Output index to file")
    parser.add_argument("--ext", type=str, default="td", help="Default file extension (if using --splitted)")
    parser.add_argument("--compat", action="store_true")
    args = parser.parse_args()

    # initialize logging
    logger.setLevel(getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)
    out_path = pathlib.Path(args.output)

    model_obj = load_model(top_level, compat=args.compat)

    metrics = {
        "n_sets": 0,
        "n_skipped": 0,
        "n_failed": 0,
        "n_success": 0,
        "skipped_sets": [],
        "failed_sets": [],
        "success_sets": [],
    }
    # preprocess model
    # print("model", model)
    artifacts = {}
    artifacts[None] = []  # used for global artifacts
    if not args.splitted:
        # content = ""
        contents = []  # Extensions need to be sorted!
        # errs = []
        settings = model_obj.settings
        llvm_version = None
        if settings:
            llvm_settings = settings.llvm
            if llvm_settings:
                llvm_state = llvm_settings.state
                if llvm_state:
                    llvm_version = llvm_state.version
        for set_name, set_def in model_obj.sets.items():
            artifacts[set_name] = []
            metrics["n_sets"] += 1
            ext_settings = set_def.settings
            if ext_settings is None:
                metrics["n_skipped"] += 1
                metrics["skipped_sets"].append(set_name)
                continue
            metrics["n_success"] += 1
            metrics["success_sets"].append(set_name)
            key, new_content = gen_riscv_isa_info_str(set_name, ext_settings=ext_settings, llvm_version=llvm_version)
            contents.append((key, new_content))
        contents = sorted(contents, key=lambda x: x[0])
        content = "\n".join([x[1] for x in contents])
        if len(content) > 0:
            with open(out_path, "w", encoding="utf-8") as f:
                f.write(content)
            if ext_settings.experimental:
                key = "riscv_isa_info_experimental"
            else:
                key = "riscv_isa_info"
            riscv_isa_info_patch = NamedPatch("llvm/lib/Support/RISCVISAInfo.cpp", key=key, src_path=out_path)
            artifacts[None].append(riscv_isa_info_patch)
    else:
        raise NotImplementedError
    if args.metrics:
        metrics_file = args.metrics
        metrics_df = pd.DataFrame({key: [val] for key, val in metrics.items()})
        metrics_df.to_csv(metrics_file, index=False)
    if args.index:
        if sum(map(len, artifacts.values())) > 0:
            global_artifacts = artifacts.get(None, [])
            set_artifacts = {key: value for key, value in artifacts.items() if key is not None}
            index_file = args.index
            write_index_yaml(index_file, global_artifacts, set_artifacts, content=True)
        else:
            logger.warning("No patches generated. No index file will be written.")


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""TODO"""

from m2isar.metamodel import arch, behav
from seal5 import model

# pylint: disable=unused-argument


def operation(self: behav.Operation, context):
    statements = []
    for stmt in self.statements:
        temp = stmt.generate(context)
        if isinstance(temp, list):
            statements.extend(temp)
        else:
            statements.append(temp)

    self.statements = statements
    return self


def binary_operation(self: behav.BinaryOperation, context):
    self.left = self.left.generate(context)
    self.right = self.right.generate(context)

    return self


def slice_operation(self: behav.SliceOperation, context):
    self.expr = self.expr.generate(context)
    self.left = self.left.generate(context)
    self.right = self.right.generate(context)

    return self


def concat_operation(self: behav.ConcatOperation, context):
    self.left = self.left.generate(context)
    self.right = self.right.generate(context)

    return self


def number_literal(self: behav.IntLiteral, context):
    return self


def int_literal(self: behav.IntLiteral, context):
    return self


def scalar_definition(self: behav.ScalarDefinition, context):
    return self


def break_(self: behav.Break, context):
    return self


def assignment(self: behav.Assignment, context):
    self.target = self.target.generate(context)
    self.expr = self.expr.generate(context)

    return self


def conditional(self: behav.Conditional, context):
    # print("conditional")
    self.conds = [x.generate(context) for x in self.conds]
    self.stmts = [x.generate(context) for x in self.stmts]

    return self


def loop(self: behav.Loop, context):
    self.cond = self.cond.generate(context)
    self.stmts = [x.generate(context) for x in self.stmts]

    return self


def ternary(self: behav.Ternary, context):
    self.cond = self.cond.generate(context)
    self.then_expr = self.then_expr.generate(context)
    self.else_expr = self.else_expr.generate(context)

    return self


def return_(self: behav.Return, context):
    if self.expr is not None:
        self.expr = self.expr.generate(context)

    return self


def unary_operation(self: behav.UnaryOperation, context):
    self.right = self.right.generate(context)

    return self


def named_reference(self: behav.NamedReference, context):
    # print("named_reference", self)
    # print("self.reference", self.reference)
    if isinstance(self.reference, arch.BitFieldDescr):
        op_name = self.reference.name
        assert op_name in context.operands
        op = context.operands[op_name]
        if isinstance(op, model.Seal5RegOperand):
            return self
        if not isinstance(op, model.Seal5ImmOperand):
            op = model.Seal5ImmOperand(op.name, op.ty, op.attributes, op.constraints)
            context.operands[op_name] = op
    # input("1111")
    return self


def indexed_reference(self: behav.IndexedReference, context):
    self.index = self.index.generate(context)

    return self


def type_conv(self: behav.TypeConv, context):
    self.expr = self.expr.generate(context)

    return self


def callable_(self: behav.Callable, context):
    self.args = [stmt.generate(context) for stmt in self.args]

    return self


def group(self: behav.Group, context):
    self.expr = self.expr.generate(context)

    return self


def procedure_call(self: behav.ProcedureCall, context):
    self.fn_args = [arg.generate(context) for arg in self.args]
    return self



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Generate Patches for RISCVRegisterInfo.td."""

import argparse
import logging
import pathlib

from seal5.index import NamedPatch, write_index_yaml

# from seal5.settings import ExtensionsSettings
from seal5.model import Seal5RegisterClass
from seal5.model_utils import load_model

from seal5.logging import Logger

logger = Logger("backends.riscv_register_info")


def write_riscv_register_info(reg):
    """Generate code for RISCVRegisterInfo.td LLVM patch (RISCVReg)."""
    # registers_template = Template(filename=str(template_dir / "registers_tablegen.mako"))

    # out_str = registers_template.render()
    out_str = ""
    if reg.is_const:
        out_str += "let isConstant = true in\n  "
    out_str += f'def {reg.name} : RISCVReg<0, "{reg.name.lower()}">;'

    return out_str


def write_riscv_register_class_info(group_name, group, group_regs):
    """Generate code for RISCVRegisterInfo.td LLVM patch (RISCVRegisterClass)."""
    # registers_template = Template(filename=str(template_dir / "registers_tablegen.mako"))

    # out_str = registers_template.render()
    # TODO: get xlen?
    reg_width = group.reg_width
    # tablegen type infer fails for types < i32?
    reg_width = max(reg_width, 32)
    VT_LOOKUP = {
        Seal5RegisterClass.GPR: ("XLenVT", reg_width),
        Seal5RegisterClass.FPR: (f"f{reg_width}", reg_width),
        Seal5RegisterClass.CSR: ("XLenVT", reg_width),
        Seal5RegisterClass.CUSTOM: (f"i{reg_width}", reg_width),
    }
    vt_name, vt_size = VT_LOOKUP.get(group.reg_class)
    vt_names = [vt_name]
    vt_sizes = [vt_size]
    vts_str = ", ".join(vt_names)
    align = max(vt_sizes)
    lo = 0
    hi = group.size - 1
    alt = group.names[0] != f"{group_name}0"
    out_str = ""
    if group.reg_class == Seal5RegisterClass.CUSTOM:
        out_str += "let isAllocatable = 0 in\n  "
    # TODO: pass lo/hi to Seal5RegisterGroup
    if alt:
        seq_str = ", ".join(group.names)
    else:
        seq_str = f'(sequence "{group_name}%u", {lo}, {hi})'
    out_str += f"def {group_name} : RISCVRegisterClass<[{vts_str}], {align}, (add {seq_str})>;"

    return out_str


def gen_riscv_register_info_str(set_def):
    """Generate full code for RISCVRegisterInfo.td patch for a set."""
    registers = set_def.registers
    register_groups = set_def.register_groups
    # print("registers", registers)
    # print("register_groups", register_groups)
    group_regs = {group_name: group.names for group_name, group in register_groups.items()}
    # print("group_regs", group_regs)
    # all_group_regs = [name for names in group_regs.values() for name in names]
    # print("all_group_regs", all_group_regs)
    ret_groups = []
    to_skip = []
    for group_name, group in register_groups.items():
        if group.reg_class == Seal5RegisterClass.GPR:
            to_skip.extend(group.names)
            continue  # Already supported
            # TODO: check size and width
        if group.reg_class == Seal5RegisterClass.FPR:
            raise NotImplementedError("Floating point registers not supported")
        elif group.reg_class == Seal5RegisterClass.CSR:
            raise NotImplementedError("CSR registers not yet supported")
        elif group.reg_class == Seal5RegisterClass.CUSTOM:
            group_regs = [registers[name] for name in group.names]
            tablegen_str = write_riscv_register_class_info(group_name, group, group_regs)
            ret_groups.append(tablegen_str)
            # raise NotImplementedError("Custom register goups not yet supported")
        else:
            raise ValueError(f"Unhandled case: {group.reg_class}")
    ret = []
    for reg_name, reg in registers.items():
        if reg_name in to_skip:
            logger.debug("Skipping register %s", reg_name)
            continue
        assert reg.reg_class not in [Seal5RegisterClass.GPR, Seal5RegisterClass.FPR, Seal5RegisterClass.CSR]
        if reg.reg_class == Seal5RegisterClass.CUSTOM:
            assert reg.size == 1
            tablegen_str = write_riscv_register_info(reg)
            ret.append(tablegen_str)
        else:
            raise ValueError(f"Unhandled case: {reg.reg_class}")
            # width = reg.width
            # TODO: use width?
    return "\n".join(ret + ret_groups)


def main():
    """Main app entrypoint."""

    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--splitted", action="store_true", help="Split per set")
    parser.add_argument("--formats", action="store_true", help="Also generate instruction formats")
    parser.add_argument("--metrics", default=None, help="Output metrics to file")
    parser.add_argument("--index", default=None, help="Output index to file")
    parser.add_argument("--ext", type=str, default="td", help="Default file extension (if using --splitted)")
    parser.add_argument("--compat", action="store_true")
    args = parser.parse_args()

    # initialize logging
    logger.setLevel(getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)
    out_path = pathlib.Path(args.output)

    model_obj = load_model(top_level, compat=args.compat)

    metrics = {
        "n_sets": 0,
        "n_skipped": 0,
        "n_failed": 0,
        "n_success": 0,
    }
    # preprocess model
    # print("model", model)
    # settings = model.get("settings", None)
    artifacts = {}
    artifacts[None] = []  # used for global artifacts
    if not args.splitted:
        content = ""
        for _, set_def in model_obj.sets.items():
            content_ = gen_riscv_register_info_str(set_def)
            if len(content_) > 0:
                content += content_
        with open(out_path, "w", encoding="utf-8") as f:
            f.write(content)
        if len(content) > 0:
            register_info_patch = NamedPatch(
                "llvm/lib/Target/RISCV/RISCVRegisterInfo.td",
                key="riscv_register_info",
                src_path=out_path,
            )
            artifacts[None].append(register_info_patch)
    else:
        raise NotImplementedError("--splitted not supported")
    if args.metrics:
        # raise NotImplementedError("--metrics not implemented")
        metrics_file = args.metrics
        with open(metrics_file, "w", encoding="utf-8") as f:
            f.write(",".join(metrics.keys()))
            f.write("\n")
            f.write(",".join(map(str, metrics.values())))
            f.write("\n")
    if args.index:
        if sum(map(len, artifacts.values())) > 0:
            global_artifacts = artifacts.get(None, [])
            set_artifacts = {key: value for key, value in artifacts.items() if key is not None}
            index_file = args.index
            write_index_yaml(index_file, global_artifacts, set_artifacts, content=True)
        else:
            logger.warning("No patches generated. No index file will be written.")


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Clean M2-ISA-R/Seal5 metamodel to .core_desc file."""

import argparse
import logging
import pathlib

import yaml

from seal5.model_utils import load_model

from seal5.logging import Logger

logger = Logger("backends.yaml_writer")

# @dataclass
# class Seal5Settings(YAMLSettings):
#     extensions: Optional[Dict[str, ExtensionsSettings]] = None
# @dataclass
# class ExtensionsSettings(YAMLSettings):
#     feature: Optional[str] = None
#     arch: Optional[str] = None
#     version: Optional[str] = None
#     experimental: Optional[bool] = None
#     vendor: Optional[bool] = None
#     instructions: Optional[List[str]] = None
#     NEW: model
#     # patches


def main():
    """Main app entrypoint."""

    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--compat", action="store_true")
    args = parser.parse_args()

    # initialize logging
    logger.setLevel(getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)
    model_name = top_level.stem
    out_path = pathlib.Path(args.output)

    model_obj = load_model(top_level, compat=args.compat)

    # preprocess model
    # print("model", model)
    data = {"extensions": {}}
    for set_name, set_def in model_obj.sets.items():
        # print("set", set_def)
        is_group_set = False
        if len(set_def.instructions) == 0:
            assert len(set_def.extension) > 0
            is_group_set = True
        set_data = {"instructions": []}
        riscv_data = {}
        if is_group_set:
            # set_data["implies"] = set_def.extension
            set_data["requires"] = set_def.extension
        else:
            riscv_data["xlen"] = set_def.xlen
            set_data["riscv"] = riscv_data
        llvm_imm_types = set()
        for instr in set_def.instructions.values():
            set_data["instructions"].append(instr.name)
            llvm_imm_types.update(instr.llvm_imm_types)
        set_data["required_imm_types"] = list(llvm_imm_types)
        data["extensions"][set_name] = set_data
    data = {"models": {model_name: data}}
    with open(out_path, "w", encoding="utf-8") as f:
        yaml.dump(data, f)


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""Clean M2-ISA-R/Seal5 metamodel to .core_desc file."""

import argparse
import logging
import pathlib
import os.path
from typing import Optional
from dataclasses import dataclass

import pandas as pd

from seal5.index import NamedPatch, write_index_yaml
from seal5.settings import IntrinsicDefn
from seal5.model_utils import load_model

from seal5.logging import Logger

logger = Logger("backends.riscv_intrinsics")


# See https://github.com/llvm-mirror/clang/blob/master/include/clang/Basic/Builtins.def
# Types:
#  v -> void
#  b -> boolean
#  c -> char
#  s -> short
#  i -> int
#  h -> half
#  f -> float
#  d -> double
#  z -> size_t
#  w -> wchar_t
#  F -> constant CFString
#  G -> id
#  H -> SEL
#  M -> struct objc_super
#  a -> __builtin_va_list
#  A -> "reference" to __builtin_va_list
#  V -> Vector, followed by the number of elements and the base type.
#  E -> ext_vector, followed by the number of elements and the base type.
#  X -> _Complex, followed by the base type.
#  Y -> ptrdiff_t
#  P -> FILE
#  J -> jmp_buf
#  SJ -> sigjmp_buf
#  K -> ucontext_t
#  p -> pid_t
#  . -> "...".  This may only occur at the end of the function list.
#
# Types may be prefixed with the following modifiers:
#  L   -> long (e.g. Li for 'long int', Ld for 'long double')
#  LL  -> long long (e.g. LLi for 'long long int', LLd for __float128)
#  LLL -> __int128_t (e.g. LLLi)
#  Z   -> int32_t (require a native 32-bit integer type on the target)
#  W   -> int64_t (require a native 64-bit integer type on the target)
#  N   -> 'int' size if target is LP64, 'L' otherwise.
#  O   -> long for OpenCL targets, long long otherwise.
#  S   -> signed
#  U   -> unsigned
#  I   -> Required to constant fold to an integer constant expression.
#
# Types may be postfixed with the following modifiers:
# * -> pointer (optionally followed by an address space number, if no address
#               space is specified than any address space will be accepted)
# & -> reference (optionally followed by an address space number)
# C -> const
# D -> volatile
#

IR_TYPE_LOOKUP_TEXT = {
    # "i32": "Li",
    "i1": "b",
    "i8": "c",
    "i16": "s",
    "i32": "Zi",
    "i64": "Wi",
    "f16": "h",
    "f32": "f",
    "f64": "d",
    # TODO: sign?
}

IR_TYPE_LOOKUP_PREFIX = {
    (None,): "U",
    (False,): "U",
    (True,): "S",
}


def ir_type_lookup_new(llvm_ty, signed: bool):
    ret = None
    if llvm_ty in ["i1"]:
        ret = "bool"
    elif llvm_ty in ["i64", "i32", "i16", "i8"]:
        sz = int(llvm_ty[1:])
        ret = f"int{sz}_t"
        if not signed:
            ret = f"u{ret}"
    elif llvm_ty in ["f64", "f32", "f16"]:
        sz = int(llvm_ty[1:])
        lookup = {
            16: "half",
            32: "float",  # single
            64: "double",
        }
        ret = lookup[sz]
    if ret is None:
        raise NotImplementedError("Unhandled llvm type: {llvm_ty}")
    return ret


IR_TYPE_LOOKUP_PAT = {
    # TODO: use f"llvm_{ir_type}_ty" instead?
    "i1": "llvm_i1_ty",
    "i8": "llvm_i8_ty",
    "i16": "llvm_i16_ty",
    "i32": "llvm_i32_ty",
    "i64": "llvm_i64_ty",
    "f16": "llvm_half_ty",
    "f32": "llvm_float_ty",
    "f64": "llvm_double_ty",
    # llvm_anyint_ty?
    # llvm_anyvector_ty?
    # llvm_ptr_ty?
}


def ir_type_to_text(ir_type: str, signed: bool = False, legacy: bool = True):
    # TODO: needs fleshing out with all likely types
    # TODO: probably needs to take into account RISC-V bit width, e.g. does "Li"
    #   means 32 bit integer on a 128-bit platform?
    if legacy:
        prefix_lookup = IR_TYPE_LOOKUP_PREFIX
        # print("prefix_lookup", prefix_lookup)
        # print("signed", signed, type(signed))
        key = (signed,)
        # print("key", key)
        prefix_found = prefix_lookup.get(key, None)
        assert prefix_found is not None, f"Unhandled prefix for ir_type '{ir_type}'"
        text_lookup = IR_TYPE_LOOKUP_TEXT
        found = text_lookup.get(ir_type, None)
        assert found is not None, f"Unhandled ir_type '{ir_type}'"
        return found
    ret = ir_type_lookup_new(ir_type, signed)
    assert ret is not None
    return ret


def build_target(arch: str, intrinsic: IntrinsicDefn):
    # Target couples intrinsic name to argument types and function behaviour
    # Start with return type if not void
    arg_str = ""
    if intrinsic.ret_type:
        arg_str += ir_type_to_text(intrinsic.ret_type, signed=intrinsic.ret_signed, legacy=True)
    for arg in intrinsic.args:
        arg_str += ir_type_to_text(arg.arg_type, signed=arg.signed, legacy=True)

    target = f'TARGET_BUILTIN(__builtin_riscv_{arch}_{intrinsic.intrinsic_name}, "{arg_str}", "nc", "{arch}")\n'
    return target


def build_target_new(
    arch: str,
    attr: str,
    intrinsic: IntrinsicDefn,
    prefix: Optional[str] = "__builtin_riscv",
    xlen: Optional[int] = None,
    has_side_effects: bool = False,
):
    # TODO: get side effects for intrinsic from instr_def?
    has_side_effects_ = has_side_effects or len(intrinsic.args) == 0
    attributes = ["NoThrow"]  # TODO: expose/use/group
    if not has_side_effects_:
        attributes.append("Const")
    if prefix != "__builtin_riscv":
        # Use: let Spellings = ["__builtin_riscv_" # NAME];
        raise NotImplementedError("Clang builtin custom prefix")
    ret_str = "void"
    if intrinsic.ret_type:
        ret_str = ir_type_to_text(intrinsic.ret_type, signed=intrinsic.ret_signed, legacy=False)
    args_str = ", ".join([ir_type_to_text(arg.arg_type, signed=arg.signed, legacy=False) for arg in intrinsic.args])
    prototype_str = f"{ret_str}({args_str})"
    features = [attr]
    if xlen is not None:
        features.append(f"{xlen}bit")
    features_str = ",".join(features)
    target = f'def {arch}_{intrinsic.intrinsic_name} : RISCVBuiltin<"{prototype_str}", "{features_str}">;'
    if len(attributes) > 0:
        attributes_str = ", ".join(attributes)
        ret = f"""let Attributes = [{attributes_str}] in {{
  {target}
}} // Attributes = [{attributes_str}]
"""
    else:
        ret = f"{target}\n"

    return ret


def ir_type_to_pattern(ir_type: str):
    # needs fleshing out with all likely types
    found = IR_TYPE_LOOKUP_PAT.get(ir_type, None)
    assert found is not None, f"Unhandled ir_type '{ir_type}'"
    return found


def build_attr(arch: str, intrinsic: IntrinsicDefn):
    # uses_mem = False  # TODO: use
    ret = f"  def int_riscv_{arch}_{intrinsic.intrinsic_name} : Intrinsic<\n    ["
    if intrinsic.ret_type:
        ret += f"{ir_type_to_pattern(intrinsic.ret_type)}"
    ret += "],\n    ["
    for idx, arg in enumerate(intrinsic.args):
        if idx:
            ret += ", "
        ret += ir_type_to_pattern(arg.arg_type)
    ret += "],\n"
    attrs = []
    no_mem = True  # TODO: expose
    if no_mem:
        attrs.append("IntrNoMem")
    if True:
        attrs.append("IntrSpeculatable")
        attrs.append("IntrWillReturn")
    has_side_effects = False  # TODO: expose
    has_side_effects_ = has_side_effects or len(intrinsic.args) == 0
    if has_side_effects_:
        attrs.append("IntrHasSideEffects")
    attrs_str = ", ".join(attrs)
    ret += f"    [{attrs_str}]>;\n"
    return ret


def build_emit(arch: str, intrinsic: IntrinsicDefn):
    emit = (
        f"  case RISCV::BI__builtin_riscv_{arch}_{intrinsic.intrinsic_name}:\n"
        f"    ID = Intrinsic::riscv_{arch}_{intrinsic.intrinsic_name};\n"
        f"    break;\n"
    )
    return emit


@dataclass
class PatchFrag:
    """Pairs patch contents to location to apply it"""

    patchee: str
    tag: str
    contents: str = ""


def main():
    """Main app entrypoint."""

    # read command line args
    parser = argparse.ArgumentParser()
    parser.add_argument("top_level", help="A .m2isarmodel or .seal5model file.")
    parser.add_argument("--log", default="info", choices=["critical", "error", "warning", "info", "debug"])
    parser.add_argument("--output", "-o", type=str, default=None)
    parser.add_argument("--splitted", action="store_true", help="Split per set")
    parser.add_argument("--formats", action="store_true", help="Also generate instruction formats")
    parser.add_argument("--metrics", default=None, help="Output metrics to file")
    parser.add_argument("--index", default=None, help="Output index to file")
    parser.add_argument("--ext", type=str, default="td", help="Default file extension (if using --splitted)")
    parser.add_argument("--ignore-failing", action="store_true", help="Do not crash in case of errors.")
    parser.add_argument("--compat", action="store_true")
    args = parser.parse_args()

    # initialize logging
    logger.setLevel(getattr(logging, args.log.upper()))

    # resolve model paths
    top_level = pathlib.Path(args.top_level)
    out_path = pathlib.Path(args.output)

    model_obj = load_model(top_level, compat=args.compat)

    metrics = {
        "n_sets": 0,
        "n_instructions": 0,
        "n_intrinsics": 0,
        "n_skipped": 0,
        "n_failed": 0,
        "n_success": 0,
        "skipped_instrinsics": [],
        "failed_intrinsics": [],
        "success_intrinsics": [],
        "skipped_instructions": [],
        "failed_instructions": [],
        "success_instructions": [],
        "skipped_sets": [],
        "failed_sets": [],
        "success_sets": [],
    }
    # preprocess model
    # print("model", model)
    artifacts = {}
    artifacts[None] = []  # used for global artifacts
    if args.splitted:
        raise NotImplementedError
    else:
        # errs = []
        settings = model_obj.settings
        llvm_version = None
        if not settings or not settings.intrinsics.intrinsics:
            logger.warning("No intrinsics configured; didn't need to invoke intrinsics writer.")
            quit()  # TODO: refactor this
        if settings:
            llvm_settings = settings.llvm
            if llvm_settings:
                llvm_state = llvm_settings.state
                if llvm_state:
                    llvm_version = llvm_state.version  # unused today, but needed very soon
                    assert llvm_version.major >= 17
        patch_frags = {
            "attr": PatchFrag(patchee="llvm/include/llvm/IR/IntrinsicsRISCV.td", tag="intrinsics_riscv"),
            "emit": PatchFrag(patchee="clang/lib/CodeGen/CGBuiltin.cpp", tag="cg_builtin"),
        }
        if llvm_version is not None and llvm_version.major < 19:
            patch_frags["target"] = PatchFrag(
                patchee="clang/include/clang/Basic/BuiltinsRISCV.def", tag="builtins_riscv"
            )
        else:
            patch_frags["target"] = PatchFrag(
                patchee="clang/include/clang/Basic/BuiltinsRISCV.td", tag="builtins_riscv"
            )
        global_riscv_settings = settings.riscv
        model_name = top_level.stem
        model_settings = settings.models.get(model_name)
        model_riscv_settings = model_settings.riscv
        riscv_settings = global_riscv_settings
        if model_riscv_settings is not None:
            riscv_settings = riscv_settings.merge(model_riscv_settings)
        for set_name, _ in model_obj.sets.items():
            artifacts[set_name] = []
            metrics["n_sets"] += 1
            # ext_settings = set_def.settings
            ext_settings = model_settings.extensions.get(set_name)
            riscv_settings_ = riscv_settings
            ext_riscv_settings = ext_settings.riscv
            if ext_riscv_settings:
                riscv_settings_ = riscv_settings.merge(ext_riscv_settings)

            if ext_settings is None or len(settings.intrinsics.intrinsics) == 0:
                metrics["n_skipped"] += 1
                metrics["skipped_sets"].append(set_name)
                continue
            assert riscv_settings_ is not None
            xlen = riscv_settings_.xlen
            assert xlen is not None
            for intrinsic in settings.intrinsics.intrinsics:
                if intrinsic.set_name is not None and intrinsic.set_name != set_name:
                    continue
                try:
                    arch_ = ext_settings.get_arch(name=set_name)
                    attr_ = ext_settings.get_attr(name=set_name)
                    if llvm_version is not None and llvm_version.major < 19:
                        patch_frags["target"].contents += build_target(arch=arch_, intrinsic=intrinsic)
                    else:
                        patch_frags["target"].contents += build_target_new(
                            arch=arch_, attr=attr_, intrinsic=intrinsic, xlen=xlen
                        )
                    patch_frags["attr"].contents += build_attr(arch=arch_, intrinsic=intrinsic)
                    patch_frags["emit"].contents += build_emit(arch=arch_, intrinsic=intrinsic)
                    metrics["n_success"] += 1
                    metrics["success_instructions"].append(intrinsic.instr_name)
                except Exception as ex:
                    logger.exception(ex)
                    metrics["n_failed"] += 1
                    metrics["failed_instructions"].append(intrinsic.instr_name)
                    # metrics["failed_intrinsics"].append(?)
            metrics["success_sets"].append(set_name)

        for id, frag in patch_frags.items():
            contents = frag.contents
            if len(contents) > 0:
                if id == "target":
                    contents = f"// {arch_}\n{contents}\n"
                elif id == "attr":
                    contents = f'let TargetPrefix = "riscv" in {{\n{contents}\n}}'
                (root, ext) = os.path.splitext(out_path)
                patch_path = root + "_" + id + ext
                with open(patch_path, "w") as f:
                    f.write(contents.strip())
                key = frag.tag
                patch = NamedPatch(frag.patchee, key=key, src_path=patch_path, content=contents.strip())
                artifacts[None].append(patch)
    if not args.ignore_failing:
        n_failed = metrics["n_failed"]
        if n_failed > 0:
            failed = metrics["failed_instructions"]
            failing_str = ", ".join(failed)
            logger.error("%s intructions failed: %s", n_failed, failing_str)
            raise RuntimeError("Abort due to errors")
    if args.metrics:
        metrics_file = args.metrics
        metrics_df = pd.DataFrame({key: [val] for key, val in metrics.items()})
        metrics_df.to_csv(metrics_file, index=False)
    if args.index:
        if sum(map(lambda x: len(x), artifacts.values())) > 0:
            global_artifacts = artifacts.get(None, [])
            set_artifacts = {key: value for key, value in artifacts.items() if key is not None}
            index_file = args.index
            write_index_yaml(index_file, global_artifacts, set_artifacts, content=True)
        else:
            logger.warning("No patches generated. No index file will be written.")


if __name__ == "__main__":
    main()



"""Classes for managing index fiels used to summarize patches."""

from typing import List, Dict, Optional
from pathlib import Path
import yaml


class Artifact:  # TODO: make abstract
    """Abstract class for artifacts used in patching procedures."""

    def __init__(
        self, dest_path: Path, src_path: Optional[Path] = None, content: Optional[str] = None, append: bool = False
    ):
        self.dest_path: Path = dest_path
        self.src_path: Optional[Path] = src_path
        self.content: Optional[str] = content
        self.append: bool = append

    def to_dict(self, content=False) -> dict:
        """Returns dict-representation of artifact."""
        return {
            key: str(value) if isinstance(value, Path) else value
            for key, value in vars(self).items()
            if content or key not in ["content"]
        }


class GitPatch(Artifact):
    """GitPatch Artifact class."""

    def __init__(
        self,
        dest_path: Path,
        src_path: Optional[Path] = None,
        content: Optional[str] = None,
        message: Optional[str] = None,
        append: bool = False,
    ):
        super().__init__(dest_path, src_path=src_path, content=content, append=append)
        self.message = "???" if message is None else message
        # TODO: context
        raise NotImplementedError


class Patch(Artifact):
    """Generic Patch artifact class."""

    def __init__(
        self, dest_path: Path, src_path: Optional[Path] = None, content: Optional[str] = None, append: bool = False
    ):
        super().__init__(dest_path, src_path=src_path, content=content, append=append)


class NamedPatch(Patch):
    """NamedPatch artifact class."""

    def __init__(
        self,
        dest_path: Path,
        key: str,
        src_path: Optional[Path] = None,
        content: Optional[Path] = None,
        append: bool = False,
    ):
        super().__init__(dest_path, src_path=src_path, content=content, append=append)
        self.key: str = key
        # self.src: str = self.out_path

    # @property
    # def out_path(self):
    #     return self.path + "." + self.key


class IndexedPatch(Patch):
    """IndexedPatch artifact class."""

    def __init__(
        self,
        dest_path: Path,
        line: int,
        src_path: Optional[Path] = None,
        content: Optional[Path] = None,
        append: bool = False,
    ):
        super().__init__(dest_path, src_path=src_path, content=content, append=append)
        # 0: start of file, -1: end of file
        self.line: int = line
        raise NotImplementedError


class RangedPatch(Patch):
    """RangedPatch artifact class."""

    def __init__(
        self,
        dest_path: Path,
        start: int,
        end: int,
        src_path: Optional[Path] = None,
        content: Optional[Path] = None,
        append: bool = False,
    ):
        super().__init__(dest_path, src_path=src_path, content=content, append=append)
        # Patch is added between matching lines
        self.start: str = start
        self.end: str = end
        raise NotImplementedError


class File(Artifact):
    """File artifact class."""

    def __init__(
        self, dest_path: Path, src_path: Optional[Path] = None, content: Optional[Path] = None, append: bool = False
    ):
        super().__init__(dest_path, src_path=src_path, content=content, append=append)
        assert self.src_path is not None or self.content is not None
        self.is_test = False


class TestFile(File):
    """File artifact class."""

    def __init__(
        self, dest_path: Path, src_path: Optional[Path] = None, content: Optional[Path] = None, append: bool = False
    ):
        super().__init__(dest_path, src_path=src_path, content=content, append=append)
        self.is_test = True


class Directory(Artifact):
    """Directory artifact class."""

    def __init__(
        self, dest_path: Path, src_path: Optional[Path] = None, content: Optional[Path] = None, append: bool = False
    ):
        assert content is None
        assert not append
        super().__init__(dest_path, src_path=src_path, content=content, append=append)


def write_index_yaml(
    out_path: Path, global_artifacts: List[Artifact], ext_artifacts: Dict[str, List[Artifact]], content=False
):
    """Export full index to YAML file."""
    extensions_yaml_data = []
    for ext, artifacts_ in ext_artifacts.items():
        extension_yaml_data = {"name": ext, "artifacts": list(map(lambda a: a.to_dict(content=content), artifacts_))}
        extensions_yaml_data.append(extension_yaml_data)
    index_yaml_data = {
        "artifacts": list(map(lambda a: a.to_dict(content=content), global_artifacts)),
        "extensions": extensions_yaml_data,
    }
    with open(out_path, "w", encoding="utf-8") as f:
        yaml.safe_dump(index_yaml_data, f)



# -*- Python -*-

# Configuration file for the 'lit' test runner.

import lit.util
import lit.formats
from lit.llvm import llvm_config

# name: The name of this test suite.
config.name = "LLVM"

# testFormat: The test format to use to interpret tests.
# config.test_format = lit.formats.ShTest(not llvm_config.use_lit_shell)
config.test_format = lit.formats.ShTest(True)

# # suffixes: A list of file extensions to treat as test files. This is overriden
# # by individual lit.local.cfg files in the test subdirectories.
config.suffixes = [".ll", ".c", ".test", ".txt", ".s", ".mir", ".yaml"]

import subprocess

# Query LLVM version
llvm_version = subprocess.check_output(["llvm-as", "--version"], text=True).split("version", 1)[1].split(" ")[1]
major_version = int(llvm_version.split(".")[0])

# Add a substitution for FileCheck
config.substitutions.append(("%llvm_major_version%", str(major_version)))



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich
"""TODO"""

import re
from enum import IntEnum, Enum, auto
from typing import Dict, List, Optional, Union

from m2isar.metamodel.arch import (
    InstructionSet,
    Instruction,
    Constant,
    Memory,
    Function,
    BaseNode,
    InstrAttribute,
    MemoryAttribute,
    BitField,
    BitVal,
    DataType,
    SizedRefOrConst,
)
from m2isar.metamodel.behav import Operation, BinaryOperation, Operator, NamedReference, IntLiteral, SliceOperation
from m2isar.metamodel.code_info import FunctionInfo
from m2isar.metamodel import M2Model

from seal5.settings import Seal5Settings, ExtensionsSettings


class Seal5InstructionSet(InstructionSet):
    """TODO."""

    def __init__(
        self,
        name: str,
        extension: "list[str]",
        constants: "dict[str, Constant]",
        memories: "dict[str, Memory]",
        functions: "dict[str, Function]",
        instructions: "dict[tuple[int, int], Instruction]",
        intrinsics: "dict[str, Seal5Intrinsic]",
        constraints: "dict[str, Seal5Constraint]",
        aliases: "dict[str, Seal5Alias]",
        registers: "dict[str, Seal5Register]",
        register_groups: "dict[str, Seal5RegisterGroup]",
    ):
        super().__init__(name, extension, constants, memories, functions, instructions)

        self.intrinsics = intrinsics
        self.constraints = constraints
        self.aliases = aliases
        self.registers = registers
        self.register_groups = register_groups
        self.settings: ExtensionsSettings = None
        self._xlen = None

    @property
    def xlen(self):
        if self._xlen is None:
            for mem_name, mem_def in self.memories.items():
                if mem_name == "X" or MemoryAttribute.IS_MAIN_REG in mem_def.attributes:
                    self._xlen = mem_def.size
                    break
        assert self._xlen is not None, "Could not determine XLEN"
        return self._xlen


class Seal5RegisterClass(IntEnum):
    GPR = 0
    FPR = 1
    CSR = 2
    CUSTOM = 3
    UNKNOWN = 4
    GPRC = 5
    # TODO: use auto()
    # TODO: GPRv2, GPRv4 // GPR32V2, GPR32V4


class Seal5Register:
    def __init__(self, name: str, size: int, width: int, signed: bool, is_const: bool, reg_class: Seal5RegisterClass):
        self.name = name
        self.size = size
        self.width = width
        self.signed = signed  # TODO: use
        self.is_const = is_const
        self.reg_class = reg_class
        # TODO: attributes

    def __repr__(self):
        return (
            f"{type(self)}({self.name}, size={self.size}, "
            f"width={self.width}, signed={self.signed}, is_const={self.is_const}"
            f"reg_class={self.reg_class})"
        )


class Seal5RegisterGroup:
    def __init__(
        self,
        names: List[str],
        reg_size: int,
        reg_width: int,
        reg_signed: bool,
        reg_is_const: bool,
        reg_class: Seal5RegisterClass,
    ):
        self.names = names
        self.reg_size = reg_size
        self.reg_width = reg_width
        self.reg_signed = reg_signed  # TODO: use
        self.reg_is_const = reg_is_const
        # TODO: drop reg-specifics?
        self.reg_class = reg_class

    def __repr__(self):
        return (
            f"{type(self)}({self.names}, size={self.size}, reg_size={self.reg_size}, reg_width={self.reg_width}, "
            f"reg_signed={self.reg_signed}, reg_is_const={self.reg_is_const}, reg_class={self.reg_class})"
        )

    @property
    def registers(self):
        return [
            Seal5Register(
                name,
                size=self.reg_size,
                width=self.reg_width,
                signed=self.reg_signed,
                is_const=self.reg_is_const,
                reg_class=self.reg_class,
            )
            for name in self.names
        ]

    @property
    def size(self):
        return len(self.names)

    def __len__(self):
        return self.size

    # TODO: allow indexing i.e. via group[12]


class Seal5Intrinsic:
    pass


class Seal5Constraint:
    stmts: List[BaseNode]
    description: Optional[str] = None

    def __init__(self, stmts, description=None):
        self.stmts = stmts
        self.description = description


class Seal5Alias:
    pass


class Seal5InstrAttribute(Enum):
    HAS_SIDE_EFFECTS = auto()
    MAY_LOAD = auto()
    MAY_STORE = auto()
    IS_TERMINATOR = auto()
    IS_BRANCH = auto()
    COMPRESSED = auto()
    USES = auto()
    DEFS = auto()
    SKIP_PATTERN_GEN = auto()
    LLVM_INSTR = auto()
    HAS_CALL = auto()
    HAS_LOOP = auto()
    OPCODE = auto()
    OPCODE_NAME = auto()
    FUNCT3 = auto()
    FUNCT7 = auto()
    ENC_FORMAT = auto()
    ENC_PATTERN = auto()
    ENC_MASK = auto()
    ENC_MATCH = auto()


class Seal5FunctionAttribute(Enum):
    INLINE = auto()


class Seal5OperandAttribute(Enum):
    IN = auto()  # or: READ (R)
    OUT = auto()  # or: WRITE (W)
    INOUT = auto()  # or: READWRITE (RW)
    UNUSED = auto()
    # LANES = auto()
    IS_REG = auto()
    IS_IMM = auto()
    TYPE = auto()
    REG_CLASS = auto()
    REG_TYPE = auto()
    IS_IMM_LEAF = auto()


class Seal5DataType(Enum):
    pass


# class Seal5OperandClass(Enum):
#     UNKNOWN = auto()
#     REG = auto()
#     IMM = auto()


class Seal5Type:
    datatype: Union[DataType, Seal5DataType] = DataType.NONE
    width: Optional[int] = None
    lanes: Optional[int] = None
    # TODO: is_vector,...

    def __init__(self, datatype, width, lanes):
        self.datatype = datatype
        self.width = width
        self.lanes = lanes

    def __repr__(self):
        sign_letter = None
        if self.datatype == DataType.U:
            sign_letter = "u"
        elif self.datatype == DataType.S:
            sign_letter = "s"
        assert sign_letter is not None
        if self.lanes is None:
            lanes = 1
        else:
            lanes = self.lanes
        if self.width is None:
            width_ = "?"
        else:
            assert self.width % lanes == 0
            width_ = self.width // lanes
        ret = f"{sign_letter}{width_}"
        if lanes > 1:
            ret = f"v{lanes}{ret}"
        return ret


class Seal5Operand:
    name: str
    ty: Seal5Type
    _attributes: Dict[Seal5OperandAttribute, List[BaseNode]] = {}
    constraints: List[Seal5Constraint] = []
    # TODO: track imm, const?
    # TODO: helpers (is_float, is_int,...)

    def __repr__(self):
        return f"{type(self)}({self.name}, ty={self.ty}, attrs={self.attributes})"

    @property
    def attributes(self):
        ret = self._attributes
        if self.ty is not None:
            # if Seal5OperandAttribute.LANES not in ret:
            #     lanes = int(self.ty.lanes)
            #     ret[Seal5OperandAttribute.LANES] = lanes
            ret[Seal5OperandAttribute.TYPE] = str(self.ty)
        return ret

    def __init__(self, name, ty, attributes, constraints):
        self.name = name
        self.ty = ty
        self._attributes = attributes
        self.constraints = constraints


class Seal5ImmOperand(Seal5Operand):
    def __init__(self, name, ty, attributes, constraints):
        super().__init__(name, ty, attributes, constraints)
        if Seal5OperandAttribute.IS_IMM not in self.attributes:
            self.attributes[Seal5OperandAttribute.IS_IMM] = []
        if Seal5OperandAttribute.OUT in self.attributes or Seal5OperandAttribute.INOUT in self.attributes:
            raise RuntimeError("Imm operands can not be outputs!")

    @property
    def attributes(self):
        ret = super().attributes
        if Seal5OperandAttribute.IS_IMM not in ret:
            ret[Seal5OperandAttribute.IS_IMM] = []
        return ret


class Seal5RegOperand(Seal5Operand):
    reg_class: Seal5RegisterClass
    reg_ty: Seal5Type

    def __init__(self, name, ty, attributes, constraints, reg_class=Seal5RegisterClass.UNKNOWN, reg_ty=None):
        super().__init__(name, ty, attributes, constraints)
        self.reg_class = reg_class
        self.reg_ty = reg_ty

    @property
    def attributes(self):
        ret = super().attributes
        if Seal5OperandAttribute.IS_REG not in ret:
            ret[Seal5OperandAttribute.IS_REG] = []
        if Seal5OperandAttribute.REG_CLASS not in ret:
            ret[Seal5OperandAttribute.REG_CLASS] = self.reg_class.name
        if Seal5OperandAttribute.REG_TYPE not in ret:
            ret[Seal5OperandAttribute.REG_TYPE] = str(self.reg_ty)
        return ret


class Seal5GPROperand(Seal5RegOperand):
    def __init__(self, name, ty, attributes, constraints, reg_ty):
        super().__init__(name, ty, attributes, constraints, Seal5RegisterClass.GPR, reg_ty)


class Seal5FPROperand(Seal5RegOperand):
    def __init__(self, name, ty, attributes, constraints, reg_ty):
        super().__init__(name, ty, attributes, constraints, Seal5RegisterClass.FPR, reg_ty)


class Seal5CSROperand(Seal5RegOperand):
    def __init__(self, name, ty, attributes, constraints, reg_ty):
        super().__init__(name, ty, attributes, constraints, Seal5RegisterClass.CSR, reg_ty)


class Seal5Instruction(Instruction):
    def __init__(
        self,
        name,
        attributes: Dict[Union[InstrAttribute, Seal5InstrAttribute], List[BaseNode]],
        encoding: "list[Union[BitField, BitVal]]",
        mnemonic: str,
        assembly: str,
        operation: Operation,
        function_info: "FunctionInfo",
        constraints: "list[Seal5Constraint]",
        operands: "dict[str, Seal5Operand]",
    ):
        del operands  # TODO: use
        super().__init__(name, attributes, encoding, mnemonic, assembly, operation, function_info)
        self.constraints = constraints
        self.operands = {}
        self._llvm_asm_str = None
        self._llvm_asm_order = None
        self._llvm_constraints = None
        self._llvm_reads = None
        self._llvm_writes = None
        self._llvm_ins_str = None
        self._llvm_outs_str = None
        self._llvm_imm_types = None
        self._process_fields()

    def get_asm_str(self):
        if self.assembly is not None:
            assert isinstance(self.assembly, str)
            return self.assembly
        assert self.operands is not None
        # print("self.operands", self.operands)
        # TODO: detect reg/imm and in/out/inout here!
        # imm_operands = [op for op in self.operands.values() if Seal5OperandAttribute.IS_IMM in op.attributes]
        # reg_operands = [op for op in self.operands.values() if Seal5OperandAttribute.IS_REG in op.attributes]
        # sorted_operands = reg_operands + imm_operands
        # assert len(sorted_operands) == len(self.operands), "Some operands are not imm or reg"
        # TODO: only sort operands if not user-provided?
        # just sort by first use?
        sorted_operands = list(self.operands.values())

        # Fallback to automatically generated asm string
        def helper(op):
            name = op.name
            attributes = op.attributes
            is_reg = Seal5OperandAttribute.IS_REG in attributes
            is_imm = Seal5OperandAttribute.IS_IMM in attributes
            if is_reg:
                name = f"name({name})"
            else:
                assert is_imm, f"Operand {name} is not a reg or imm"
            name = "{" + name + "}"
            # print("name", name)
            return name

        asm_str = ", ".join([helper(op) for op in sorted_operands])
        # print("asm_str", asm_str)
        # input(">")
        self.assembly = asm_str
        return self.assembly

    def _process_fields(self):
        for field_name, field in self.fields.items():
            if field_name in self.operands:
                continue
            width = field.size
            datatype = field.data_type
            lanes = None
            ty = Seal5Type(width=width, datatype=datatype, lanes=lanes)
            # op_attrs = {Seal5OperandAttribute.IN: []}
            op_attrs = {}
            # check for fixed bits
            temp = [False] * width
            for enc in self.encoding:
                if isinstance(enc, BitField):
                    if enc.name == field_name:
                        rng = enc.range
                        assert rng.lower <= rng.upper
                        for pos in range(rng.lower, rng.upper + 1):
                            assert pos < len(temp)
                            temp[pos] = True
            temp = [pos for pos, val in enumerate(temp) if not val]
            temp2 = []
            cur = None
            for pos in temp:
                if cur is None:
                    cur = [pos, pos]
                else:
                    if pos == cur[1] + 1:
                        cur[1] = pos
                    else:
                        temp2.append(f"{cur[0]}:{cur[1]}")
                        cur = [pos, pos]
            if cur is not None:
                temp2.append(f"{cur[0]}:{cur[1]}")
            temp = temp2
            # input("pp")
            constraints = []
            for pos in temp:
                lower, upper = pos.split(":", 1)
                lower = int(lower)
                upper = int(upper)
                sz = upper - lower + 1
                stmt = BinaryOperation(
                    SliceOperation(
                        NamedReference(SizedRefOrConst(field_name, sz)), IntLiteral(upper), IntLiteral(lower)
                    ),
                    Operator("=="),
                    IntLiteral(0),
                )
                constraint = Seal5Constraint([stmt])
                constraints.append(constraint)
            if len(constraints) > 0:
                # input("eee")
                pass
            # if "imm" in field_name:
            #     cls = Seal5ImmOperand
            # elif field_name in ["rd", "rs1", "rs2", "rs3"]:
            #     cls = Seal5RegOperand
            cls = Seal5Operand
            op = cls(field_name, ty, op_attrs, constraints)
            self.operands[field_name] = op
        # Test:
        # self.attributes[Seal5InstrAttribute.MAY_LOAD] = []

    def _llvm_check_operands(self):
        asm_order = self.llvm_asm_order
        operands = self.operands
        # check that number of operands is equal
        assert len(asm_order) == len(
            operands
        ), f"Number of operands does not match ({asm_order} vs. {list(operands.keys())})"
        # check that order of operands matches asm syntax
        # for op_idx, op_name in enumerate(operands.keys()):
        #     asm_idx = asm_order.index(f"${op_name}")
        #     assert asm_idx == op_idx, "Order of asm operands does not match CDSL operands"

    def _llvm_process_operands(self):
        operands = self.operands
        reads = []
        writes = []
        constraints = []
        self._llvm_check_operands()
        imm_types = set()
        for op_name, op in operands.items():
            if len(op.constraints) > 0:
                raise NotImplementedError
            if Seal5OperandAttribute.IS_REG in op.attributes:
                assert Seal5OperandAttribute.REG_CLASS in op.attributes
                cls = op.attributes[Seal5OperandAttribute.REG_CLASS]
                assert cls in ["GPR", "GPRC"]
                pre = cls
            elif Seal5OperandAttribute.IS_IMM in op.attributes:
                assert Seal5OperandAttribute.TYPE in op.attributes
                ty = op.attributes[Seal5OperandAttribute.TYPE]
                assert ty[0] in ["u", "s"]
                sz = int(ty[1:])
                pre = f"{ty[0]}imm{sz}"
                imm_types.add(pre)
                # TODO: handle lsb0, lsb00,...
                # TODO: annotate operands via attributes

            if Seal5OperandAttribute.INOUT in op.attributes or (
                Seal5OperandAttribute.OUT in op.attributes and Seal5OperandAttribute.IN in op.attributes
            ):
                op_str2 = f"{pre}:${op_name}_wb"
                writes.append(op_str2)
                op_str = f"{pre}:${op_name}"
                reads.append(op_str)
                constraint = f"${op_name} = ${op_name}_wb"
                constraints.append(constraint)

            elif Seal5OperandAttribute.OUT in op.attributes:
                op_str = f"{pre}:${op_name}"
                writes.append(op_str)
            elif Seal5OperandAttribute.IN in op.attributes:
                op_str = f"{pre}:${op_name}"
                reads.append(op_str)
        self._llvm_constraints = constraints
        self._llvm_reads = reads
        self._llvm_writes = writes
        self._llvm_imm_types = imm_types

    def _llvm_process_assembly(self):
        asm_str = self.get_asm_str()
        assert asm_str is not None
        asm_str = re.sub(r"name\(([a-zA-Z0-9_\+]+)\)", r"\g<1>", asm_str)
        asm_str = re.sub(r"{([a-zA-Z0-9_\+]+):[#0-9a-zA-Z\._]+}", r"{\g<1>}", asm_str)
        asm_str = re.sub(r"{([a-zA-Z0-9_\+]+)}", r"$\g<1>", asm_str)
        # remove offsets
        asm_str = re.sub(r"[0-9]+\+([a-zA-Z0-9]+)", r"\g<1>", asm_str)
        asm_str = re.sub(r"([a-zA-Z0-9]+)\+[0-9]+", r"\g<1>", asm_str)
        asm_order = re.compile(r"(\$[a-zA-Z0-9_]+)").findall(asm_str)
        for op in asm_order:
            if f"{op}(" in asm_str or f"{op})" in asm_str or f"{op}!" in asm_str or f"!{op}" in asm_str:
                asm_str = asm_str.replace(op, "${" + op[1:] + "}")
        self._llvm_asm_str = asm_str
        self._llvm_asm_order = asm_order

    @property
    def llvm_asm_str(self):
        if self._llvm_asm_str is None:
            self._llvm_process_assembly()
        return self._llvm_asm_str

    @property
    def llvm_asm_order(self):
        if self._llvm_asm_order is None:
            self._llvm_process_assembly()
        return self._llvm_asm_order

    @property
    def llvm_constraints(self):
        if self._llvm_constraints is None:
            self._llvm_process_operands()
        return self._llvm_constraints

    @property
    def llvm_reads(self):
        if self._llvm_reads is None:
            self._llvm_process_operands()
        return self._llvm_reads

    @property
    def llvm_writes(self):
        if self._llvm_writes is None:
            self._llvm_process_operands()
        return self._llvm_writes

    @property
    def llvm_imm_types(self):
        if self._llvm_imm_types is None:
            self._llvm_process_operands()
        return self._llvm_imm_types

    @property
    def llvm_ins_str(self):
        if self._llvm_ins_str is None:
            reads = self.llvm_reads
            reads_ = [(x.split(":", 1)[1] if ":" in x else x) for x in reads]
            ins_str = ", ".join([reads[reads_.index(x)] for x in self._llvm_asm_order if x in reads_])
            if len(ins_str) == 0:
                assert len(reads) == 0
            self._llvm_ins_str = ins_str
        return self._llvm_ins_str

    @property
    def llvm_outs_str(self):
        if self._llvm_outs_str is None:
            writes = self.llvm_writes
            writes_ = [(x.split(":", 1)[1] if ":" in x else x).replace("_wb", "") for x in writes]
            outs_str = ", ".join([writes[writes_.index(x)] for x in self._llvm_asm_order if x in writes_])
            if len(outs_str) == 0:
                assert len(writes) == 0
            self._llvm_outs_str = outs_str
        return self._llvm_outs_str

    @property
    def llvm_attributes(self):
        attrs = {}
        if Seal5InstrAttribute.HAS_SIDE_EFFECTS in self.attributes:
            attrs["hasSideEffects"] = 1
        else:
            attrs["hasSideEffects"] = 0
        if Seal5InstrAttribute.MAY_LOAD in self.attributes:
            attrs["mayLoad"] = 1
        else:
            attrs["mayLoad"] = 0
        if Seal5InstrAttribute.MAY_STORE in self.attributes:
            attrs["mayStore"] = 1
        else:
            attrs["mayStore"] = 0
        if Seal5InstrAttribute.IS_TERMINATOR in self.attributes:
            attrs["isTerminator"] = 1
        else:
            attrs["isTerminator"] = 0
        uses = self.attributes.get(Seal5InstrAttribute.USES, [])
        if uses:
            attrs["Uses"] = "[" + ", ".join(uses) + "]"
        defs = self.attributes.get(Seal5InstrAttribute.DEFS, [])
        if defs:
            attrs["Defs"] = "[" + ", ".join(defs) + "]"
        return attrs

    def llvm_get_compressed_pat(self, set_def):
        # TODO: map metamodel instr names to llvm instr names
        uncompressed_instr = self.attributes.get(Seal5InstrAttribute.COMPRESSED, None)
        if uncompressed_instr is None or len(uncompressed_instr) == 0:
            return None
        if isinstance(uncompressed_instr, list):
            uncompressed_instr = uncompressed_instr[0]
        compressed_instr = self.name
        uncompressed_instr = uncompressed_instr.value.strip()
        uncompressed = None
        assert set_def is not None
        for _, instr_def in set_def.instructions.items():
            if instr_def.name == uncompressed_instr:
                uncompressed = instr_def
                break
        assert uncompressed is not None, f"Could not find instr {uncompressed_instr} in set {set_def.name}"
        # TODO: rs1_wb vs. rd_wb?
        uncompressed_asm_order_ = None
        assert len(uncompressed.llvm_constraints) == 0
        if len(self.llvm_asm_order) == len(uncompressed.llvm_asm_order):
            raise NotImplementedError
        if len(self.llvm_asm_order) < len(uncompressed.llvm_asm_order):
            assert len(self.llvm_asm_order) == (len(uncompressed.llvm_asm_order) - 1)
            missing = set(uncompressed.llvm_asm_order) - set(self.llvm_asm_order)
            assert len(missing) == 1
            missing = list(missing)[0]
            assert len(self.llvm_constraints) > 0
            wb_reg = None
            for constr in self.llvm_constraints:
                m = re.compile(r"(\$[a-zA-Z0-9]*)\s*=\s*(\$[a-zA-Z0-9]*)_wb").match(constr)
                if m:
                    x, y = m.groups()
                    if x == y:
                        wb_reg = x

            def replace_helper(x):
                if x == missing:
                    return wb_reg
                # if x == wb_reg:
                #     return f"{wb_reg}_wb"
                return x

            uncompressed_asm_order_ = [replace_helper(x) for x in uncompressed.llvm_asm_order]
        else:
            raise RuntimeError("Compressed instr has more args than uncompressed one")
        assert uncompressed_asm_order_ is not None
        # compressed_args = "GPRC:$rs1, GPRC:$rs2"

        def args_helper(asm_order, ins_str, outs_str):
            ret = []
            for x in asm_order:
                for y in outs_str.split(", ") + ins_str.split(", "):
                    y_ = y.split(":", 1)[-1]
                    if y_ == x:
                        ret.append(y)
                        break
            assert len(ret) == len(asm_order)
            return ", ".join(ret)

        compressed_args = args_helper(self.llvm_asm_order, self.llvm_ins_str, self.llvm_outs_str)
        uncompressed_args = args_helper(uncompressed_asm_order_, self.llvm_ins_str, self.llvm_outs_str)
        # uncompressed_args = "GPRC:$rs1, GPRC:$rs1, GPRC:$rs2"
        uncompressed_str = f"{uncompressed_instr} {uncompressed_args}"
        compressed_str = f"{compressed_instr} {compressed_args}"
        ret = f"def : CompressPat<({uncompressed_str}), ({compressed_str})>;"

        # raise NotImplementedError
        return ret


SEAL5_METAMODEL_VERSION = 2


class Seal5Model(M2Model):
    settings: Seal5Settings = None



# SPDX-License-Identifier: Apache-2.0
#
# This file is part of the M2-ISA-R project: https://github.com/tum-ei-eda/M2-ISA-R
#
# Copyright (C) 2022
# Chair of Electrical Design Automation
# Technical University of Munich

"""TODO"""

from m2isar.metamodel import arch, behav
from seal5 import model

# pylint: disable=unused-argument


def operation(self: behav.Operation, context):
    statements = []
    for stmt in self.statements:
        temp = stmt.generate(context)
        if isinstance(temp, list):
            statements.extend(temp)
        else:
            statements.append(temp)

    self.statements = statements
    return self


def binary_operation(self: behav.BinaryOperation, context):
    self.left = self.left.generate(context)
    self.right = self.right.generate(context)

    return self


def slice_operation(self: behav.SliceOperation, context):
    self.expr = self.expr.generate(context)
    self.left = self.left.generate(context)
    self.right = self.right.generate(context)

    return self


def concat_operation(self: behav.ConcatOperation, context):
    self.left = self.left.generate(context)
    self.right = self.right.generate(context)

    return self


def number_literal(self: behav.IntLiteral, context):
    return self


def int_literal(self: behav.IntLiteral, context):
    return self


def scalar_definition(self: behav.ScalarDefinition, context):
    return self


def break_(self: behav.Break, context):
    return self


def assignment(self: behav.Assignment, context):
    self.target = self.target.generate(context)
    self.expr = self.expr.generate(context)

    return self


def conditional(self: behav.Conditional, context):
    # print("conditional")
    self.conds = [x.generate(context) for x in self.conds]
    self.stmts = [x.generate(context) for x in self.stmts]

    return self


def loop(self: behav.Loop, context):
    self.cond = self.cond.generate(context)
    self.stmts = [x.generate(context) for x in self.stmts]

    return self


def ternary(self: behav.Ternary, context):
    self.cond = self.cond.generate(context)
    self.then_expr = self.then_expr.generate(context)
    self.else_expr = self.else_expr.generate(context)

    return self


def return_(self: behav.Return, context):
    if self.expr is not None:
        self.expr = self.expr.generate(context)

    return self


def unary_operation(self: behav.UnaryOperation, context):
    self.right = self.right.generate(context)

    return self


def named_reference(self: behav.NamedReference, context):
    # if isinstance(self.reference, arch.BitFieldDescr):
    # raise NotImplementedError
    return self


def indexed_reference(self: behav.IndexedReference, context):
    # print("indexed_reference", self)
    # print("self.reference", self.reference)
    # print("self.index", self.index)
    if isinstance(self.reference, arch.Memory):
        mem_name = self.reference.name
        mem = self.reference
        mem_size = mem.size
        mem_lanes = 1
        index = self.index
        offset = 0
        if isinstance(index, behav.BinaryOperation):  # offset?
            # print("binop!")
            if index.op.value == "+":
                # print("offset!")
                if isinstance(index.left, behav.NamedReference):
                    # print("lhs named")
                    if isinstance(index.right, behav.IntLiteral):
                        # print("rhs const")
                        offset = index.right.value
                        index = index.left
                elif isinstance(index.right, behav.NamedReference):
                    # print("rhs named")
                    if isinstance(index.left, behav.IntLiteral):
                        # print("lhs const")
                        offset = index.left.value
                        index = index.right
        # print("index", index)
        # print("offset", offset)
        if isinstance(index, behav.NamedReference):
            if isinstance(index.reference, arch.BitFieldDescr):
                op_name = index.reference.name
                assert op_name in context.operands
                op = context.operands[op_name]
                if mem_name == "X":
                    if offset == 0:
                        reg_class = model.Seal5RegisterClass.GPR
                    elif offset == 8:
                        reg_class = model.Seal5RegisterClass.GPRC
                    else:
                        raise NotImplementedError(f"GPR with offset {offset}")
                elif mem_name == "F":
                    assert offset == 0
                    reg_class = model.Seal5RegisterClass.FPR
                elif mem_name == "CSR":
                    assert offset == 0
                    reg_class = model.Seal5RegisterClass.CSR
                else:
                    assert offset == 0  # TODO: write offset to Operand class?
                    reg_class = model.Seal5RegisterClass.UNKNOWN
                if not isinstance(op, model.Seal5RegOperand):
                    op = model.Seal5RegOperand(op.name, op.ty, op.attributes, op.constraints, reg_class=reg_class)
                reg_ty = model.Seal5Type(arch.DataType.U, mem_size, mem_lanes)
                op.reg_ty = reg_ty
                context.operands[op_name] = op
    # input("1111")
    self.index = self.index.generate(context)

    return self


def type_conv(self: behav.TypeConv, context):
    self.expr = self.expr.generate(context)

    return self


def callable_(self: behav.Callable, context):
    self.args = [stmt.generate(context) for stmt in self.args]

    return self


def group(self: behav.Group, context):
    self.expr = self.expr.generate(context)

    return self


def procedure_call(self: behav.ProcedureCall, context):
    self.fn_args = [arg.generate(context) for arg in self.args]
    return self
